from collections import defaultdict

from robot.api.parsing import Token
from robot.parsing.model import Statement

from robotidy.disablers import skip_section_if_disabled
from robotidy.transformers import Transformer
from robotidy.utils import is_blank_multiline, left_align, round_to_four, tokens_by_lines


class AlignSettingsSection(Transformer):
    """
    Align statements in ``*** Settings ***`` section to columns.

    Following code:

    ```robotframework
    *** Settings ***
    Library      SeleniumLibrary
    Library   Mylibrary.py
    Variables  variables.py
    Test Timeout  1 min
        # this should be left aligned
    ```

    will be transformed to:

    ```robotframework
    *** Settings ***
    Library         SeleniumLibrary
    Library         Mylibrary.py
    Variables       variables.py
    Test Timeout    1 min
    # this should be left aligned
    ```

    You can configure how many columns should be aligned to longest token in given column. The remaining columns
    will use fixed length separator length ``--spacecount``. By default only first two columns are aligned.
    To align first three columns:

    ```
    robotidy --transform AlignSettingsSection:up_to_column=3
    ```

    To align all columns set ``up_to_column`` to 0.

    Arguments inside keywords in Suite Setup, Suite Teardown, Test Setup and Test Teardown are indented by additional
    ``argument_indent`` (default ``4``) spaces:

    ```robotframework
    *** Settings ***
    Suite Setup         Start Session
    ...                     host=${IPADDRESS}
    ...                     user=${USERNAME}
    ...                     password=${PASSWORD}
    Suite Teardown      Close Session
    ```

    To disable it configure ``argument_indent`` with ``0``.

    Supports global formatting param ``--spacecount`` (for columns with fixed length).
    """

    TOKENS_WITH_ARGUMENTS = {
        Token.SUITE_SETUP,
        Token.SUITE_TEARDOWN,
        Token.TEST_SETUP,
        Token.TEST_TEARDOWN,
        Token.LIBRARY,
        Token.VARIABLES,
    }

    def __init__(self, up_to_column: int = 2, argument_indent: int = 4, min_width: int = None):
        super().__init__()
        self.up_to_column = up_to_column - 1
        self.argument_indent = argument_indent
        self.min_width = min_width

    @skip_section_if_disabled
    def visit_SettingSection(self, node):  # noqa
        statements = []
        for child in node.body:
            if self.disablers.is_node_disabled(child):
                statements.append(child)
            elif child.type in (Token.EOL, Token.COMMENT):
                statements.append(left_align(child))
            else:
                statements.append(list(tokens_by_lines(child)))
        nodes_to_be_aligned = [st for st in statements if isinstance(st, list)]
        if not nodes_to_be_aligned:
            return node
        look_up = self.create_look_up(nodes_to_be_aligned)  # for every col find longest value
        node.body = self.align_rows(statements, look_up)
        return node

    def should_indent_arguments(self, statement):
        statement_type = statement[0][0].type
        is_library = statement_type == Token.LIBRARY
        if is_library:
            return is_library, True
        return is_library, statement_type in self.TOKENS_WITH_ARGUMENTS

    def align_rows(self, statements, look_up):
        aligned_statements = []
        for st in statements:
            if not isinstance(st, list):
                aligned_statements.append(st)
                continue
            is_library, indent_args = self.should_indent_arguments(st)
            aligned_statement = []
            for line in st:
                if is_blank_multiline(line):
                    line[-1].value = line[-1].value.lstrip(" \t")  # normalize eol from '  \n' to '\n'
                    aligned_statement.extend(line)
                    continue
                indent_arg = indent_args and line[0].type == Token.CONTINUATION
                if indent_arg and is_library:
                    indent_arg = all(token.type != Token.WITH_NAME for token in line)
                up_to = self.up_to_column if self.up_to_column != -1 else len(line) - 2
                for index, token in enumerate(line[:-2]):
                    aligned_statement.append(token)
                    separator = self.calc_separator(index, up_to, indent_arg, token, look_up)
                    aligned_statement.append(Token(Token.SEPARATOR, separator))
                last_token = line[-2]
                # remove leading whitespace before token
                last_token.value = last_token.value.strip() if last_token.value else last_token.value
                aligned_statement.append(last_token)
                aligned_statement.append(line[-1])  # eol
            aligned_statements.append(Statement.from_tokens(aligned_statement))
        return aligned_statements

    def calc_separator(self, index, up_to, indent_arg, token, look_up):
        if index < up_to:
            if self.min_width:
                return max(self.min_width - len(token.value), self.formatting_config.space_count) * " "
            arg_indent = self.argument_indent if indent_arg else 0
            if indent_arg and index != 0:
                return (
                    max(
                        (look_up[index] - len(token.value) - arg_indent + 4),
                        self.formatting_config.space_count,
                    )
                    * " "
                )
            else:
                return (look_up[index] - len(token.value) + arg_indent + 4) * " "
        else:
            return self.formatting_config.space_count * " "

    def create_look_up(self, statements):
        look_up = defaultdict(int)
        for st in statements:
            is_doc = st[0][0].type == Token.DOCUMENTATION
            for line in st:
                if is_doc:
                    up_to = 1
                elif self.up_to_column != -1:
                    up_to = self.up_to_column
                else:
                    up_to = len(line)
                for index, token in enumerate(line[:up_to]):
                    look_up[index] = max(look_up[index], len(token.value))
        return {index: round_to_four(length) for index, length in look_up.items()}



from functools import lru_cache
import re
import sys
from typing import Dict, Optional, List, Tuple

from robocorp_ls_core.lsp import (
    DiagnosticSeverity,
    DiagnosticTag,
    ICustomDiagnosticDataUndefinedKeywordTypedDict,
    ICustomDiagnosticDataUndefinedResourceTypedDict,
    ICustomDiagnosticDataUndefinedVarImportTypedDict,
    ICustomDiagnosticDataUndefinedLibraryTypedDict,
)
from robocorp_ls_core.protocols import check_implements
from robocorp_ls_core.robotframework_log import get_logger
from robotframework_ls.impl.ast_utils import MAX_ERRORS
from robotframework_ls.impl.protocols import (
    IKeywordFound,
    IKeywordCollector,
    ICompletionContext,
    ILibraryDoc,
    INode,
    IVariableFound,
    AbstractVariablesCollector,
    VariableKind,
    AbstractKeywordCollector,
)
from robotframework_ls.impl.robot_lsp_constants import (
    OPTION_ROBOT_LINT_VARIABLES,
    OPTION_ROBOT_LINT_IGNORE_VARIABLES,
    OPTION_ROBOT_LINT_IGNORE_ENVIRONMENT_VARIABLES,
)
from robotframework_ls.impl.robot_constants import STDLIBS_LOWER
import typing


log = get_logger(__name__)


class _KeywordContainer(object):
    def __init__(self) -> None:
        self._name_to_keywords: Dict[str, List[IKeywordFound]] = {}
        self._names_with_variables: Dict[str, List[IKeywordFound]] = {}

    def add_keyword(self, keyword_found: IKeywordFound) -> None:
        from robotframework_ls.impl.text_utilities import normalize_robot_name

        normalized_name = normalize_robot_name(keyword_found.keyword_name)
        lst = self._name_to_keywords.get(normalized_name)
        if lst is None:
            lst = self._name_to_keywords[normalized_name] = []
        lst.append(keyword_found)

        if "{" in normalized_name:
            lst = self._names_with_variables.get(normalized_name)
            if lst is None:
                lst = self._names_with_variables[normalized_name] = []
            lst.append(keyword_found)

    def get_keywords(self, normalized_keyword_name: str) -> List[IKeywordFound]:
        from robotframework_ls.impl.text_utilities import matches_name_with_variables

        keyword_found = self._name_to_keywords.get(normalized_keyword_name)

        ret = []
        if keyword_found is not None:
            ret.extend(keyword_found)

        # We do not have an exact match, still, we need to check if we may
        # have a match in keywords that accept variables.
        for name, keyword_found in self._names_with_variables.items():
            if matches_name_with_variables(normalized_keyword_name, name):
                ret.extend(keyword_found)

        return ret


class _VariablesCollector(AbstractVariablesCollector):
    def __init__(self, on_unresolved_variable_import):
        self._variables_collected: Dict[str, List[IVariableFound]] = {}
        self._template_variables_collected: List[Tuple[str, IVariableFound]] = []
        self.on_unresolved_variable_import = on_unresolved_variable_import

        self._env_variables_collected: Dict[str, IVariableFound] = {}

    def accepts(self, variable_name: str) -> bool:
        return True

    def on_env_variable(self, variable_found: IVariableFound):
        upper = variable_found.variable_name.upper()
        self._env_variables_collected[upper] = variable_found

    def on_variable(self, variable_found: IVariableFound):
        from robotframework_ls.impl.text_utilities import normalize_robot_name

        normalized = normalize_robot_name(variable_found.variable_name)
        if "{" in normalized:
            self._template_variables_collected.append((normalized, variable_found))

        lst = self._variables_collected.get(normalized)
        if lst is None:
            lst = self._variables_collected[normalized] = []
        lst.append(variable_found)

    def contains_env_variable(self, variable_name_upper: str) -> bool:
        return variable_name_upper in self._env_variables_collected

    def contains_variable(
        self, variable_name: str, var_line: int, var_col_offset: int
    ) -> bool:
        from robotframework_ls.impl.text_utilities import matches_name_with_variables

        variables_found: Optional[List[IVariableFound]] = self._variables_collected.get(
            variable_name
        )
        if variables_found:
            for v in variables_found:
                if v.lineno < var_line:
                    return True

                if (
                    v.lineno == var_line
                    and v.variable_kind == VariableKind.ARGUMENT
                    and v.col_offset < var_col_offset
                ):
                    return True

        for template_var, v in self._template_variables_collected:
            if v.lineno < var_line:
                if matches_name_with_variables(variable_name, template_var):
                    return True

        return False


class _AnalysisKeywordsCollector(object):
    def __init__(
        self,
        on_unresolved_library,
        on_unresolved_resource,
        on_resolved_library,
    ):
        self._keywords_container = _KeywordContainer()
        self._resource_name_to_keywords_container = {}
        self._library_name_to_keywords_container = {}

        self.on_unresolved_library = on_unresolved_library
        self.on_unresolved_resource = on_unresolved_resource
        self.on_resolved_library = on_resolved_library

    def accepts(self, keyword_name):
        return True

    def on_keyword(self, keyword_found: IKeywordFound):
        from robotframework_ls.impl.text_utilities import normalize_robot_name

        # Note: Even if something is imported 'WITH NAME', it's still added
        # to the global scope (there's just an additional reference with the
        # new name).
        self._keywords_container.add_keyword(keyword_found)
        library_name = keyword_found.library_name
        library_alias = keyword_found.library_alias
        resource_name = keyword_found.resource_name

        if library_name:
            if library_alias:
                name = normalize_robot_name(library_alias)
            else:
                name = normalize_robot_name(library_name)
            dct = self._library_name_to_keywords_container
        elif resource_name:
            name = normalize_robot_name(resource_name)
            dct = self._resource_name_to_keywords_container
        else:
            log.info(
                "No library name nor resource name for keyword: %s",
                keyword_found.keyword_name,
            )
            return

        keyword_container = dct.get(name)
        if keyword_container is None:
            keyword_container = dct[name] = _KeywordContainer()

        keyword_container.add_keyword(keyword_found)

    def get_keywords(self, normalized_keyword_name: str) -> List[IKeywordFound]:
        from robotframework_ls.impl import text_utilities

        ret = []

        # Note: the name could be something as `alias.keywordname` or
        # 'libraryname.keywordname`. In this case, we need to verify if there's
        # a library/alias with that specific name.
        for name, remainder in text_utilities.iter_dotted_names(
            normalized_keyword_name
        ):
            if not name or not remainder:
                continue
            containers = []
            keywords_container = self._resource_name_to_keywords_container.get(name)
            if keywords_container:
                containers.append(keywords_container)
            keywords_container = self._library_name_to_keywords_container.get(name)
            if keywords_container:
                containers.append(keywords_container)

            for keywords_container in containers:
                ret.extend(keywords_container.get_keywords(remainder))

        if not ret:
            # Finding with a dotted name has higher priority over finding it
            # without the qualifier.
            ret.extend(self._keywords_container.get_keywords(normalized_keyword_name))

        return ret

    def __typecheckself__(self) -> None:
        _: IKeywordCollector = check_implements(self)


def collect_analysis_errors(initial_completion_context):
    from robotframework_ls.impl import ast_utils
    from robotframework_ls.impl.ast_utils import create_error_from_node
    from robotframework_ls.impl.collect_keywords import collect_keywords
    from robotframework_ls.impl.text_utilities import normalize_robot_name
    from robotframework_ls.impl.text_utilities import contains_variable_text
    from robotframework_ls.impl.keyword_argument_analysis import (
        UsageInfoForKeywordArgumentAnalysis,
    )
    from robot.api import Token

    errors = []
    config = initial_completion_context.config

    def on_resolved_library(
        completion_context: ICompletionContext,
        library_node: Optional[INode],
        library_doc: ILibraryDoc,
    ):
        if library_node is None:
            return

        from robotframework_ls.impl.robot_lsp_constants import (
            OPTION_ROBOT_LINT_KEYWORD_CALL_ARGUMENTS,
        )

        if config is not None and not config.get_setting(
            OPTION_ROBOT_LINT_KEYWORD_CALL_ARGUMENTS, bool, True
        ):
            return

        doc = completion_context.doc
        if not doc or doc.uri != initial_completion_context.doc.uri:
            return

        from robotframework_ls.impl.keyword_argument_analysis import (
            KeywordArgumentAnalysis,
        )

        library_args = []
        if library_doc.inits:
            keyword_doc = library_doc.inits[0]
            library_args = keyword_doc.args

        # Ok, we found the keyword, let's check if the arguments are correct.
        keyword_argument_analysis = KeywordArgumentAnalysis(library_args)

        name_token = library_node.get_token(Token.NAME)
        if name_token is not None:
            for error in keyword_argument_analysis.collect_keyword_usage_errors(
                UsageInfoForKeywordArgumentAnalysis(library_node, name_token)
            ):
                errors.append(error)

    def on_unresolved_library(
        completion_context: ICompletionContext,
        library_name: str,
        lineno: int,
        end_lineno: int,
        col_offset: int,
        end_col_offset: int,
        error_msg: Optional[str],
        resolved_name: str,
    ):
        from robotframework_ls.impl.robot_lsp_constants import (
            OPTION_ROBOT_LINT_UNDEFINED_LIBRARIES,
        )
        from robotframework_ls.impl.robot_version import get_robot_major_version

        if config is not None and not config.get_setting(
            OPTION_ROBOT_LINT_UNDEFINED_LIBRARIES, bool, True
        ):
            return

        doc = completion_context.doc
        if doc and doc.uri == initial_completion_context.doc.uri:
            start = (lineno - 1, col_offset)
            end = (end_lineno - 1, end_col_offset)

            additional = ""
            if error_msg is None:
                error_msg = ""
            else:
                if "expected" in error_msg:
                    if re.search(r"expected\s+(.*)\s+argument(s)?,\s+got", error_msg):
                        additional = f'\nConsider using default arguments in the {library_name} constructor and\ncalling the "Robot Framework: Clear caches and restart" action or\nadding "{library_name}" to the "robot.libraries.libdoc.needsArgs"\nsetting to pass the typed arguments when generating the libspec.'

                if not additional and "Importing" in error_msg:
                    if get_robot_major_version() <= 3:
                        pattern = r"Importing\s+test\s+library(.*)failed"
                    else:
                        pattern = r"Importing\s+library(.*)failed"
                    if re.search(pattern, error_msg):
                        additional = f'\nConsider adding the needed paths to the "robot.pythonpath" setting\nand calling the "Robot Framework: Clear caches and restart" action.'

                    if "{" in library_name and resolved_name:
                        error_msg += f"\nNote: resolved name: {resolved_name}"

            error = ast_utils.Error(
                f"Unresolved library: {library_name}.{error_msg}{additional}".strip(),
                start,
                end,
            )
            undefined_var_import_data: ICustomDiagnosticDataUndefinedLibraryTypedDict = {
                "kind": "undefined_library",
                "name": library_name,
                "resolved_name": resolved_name,
            }
            error.data = undefined_var_import_data

            errors.append(error)

    def on_unresolved_resource(
        completion_context: ICompletionContext,
        resource_name: str,
        lineno: int,
        end_lineno: int,
        col_offset: int,
        end_col_offset: int,
        error_msg: Optional[str],
        resolved_name: str,
    ):
        from robotframework_ls.impl.robot_lsp_constants import (
            OPTION_ROBOT_LINT_UNDEFINED_RESOURCES,
        )

        if config is not None and not config.get_setting(
            OPTION_ROBOT_LINT_UNDEFINED_RESOURCES, bool, True
        ):
            return

        doc = completion_context.doc
        if doc and doc.uri == initial_completion_context.doc.uri:
            start = (lineno - 1, col_offset)
            end = (end_lineno - 1, end_col_offset)
            if not error_msg:
                error_msg = f"Unresolved resource: {resource_name}"
                if "{" in resource_name and resolved_name:
                    error_msg += f"\nNote: resolved name: {resolved_name}"

            error = ast_utils.Error(error_msg, start, end)
            undefined_resource_data: ICustomDiagnosticDataUndefinedResourceTypedDict = {
                "kind": "undefined_resource",
                "name": resource_name,
                "resolved_name": resolved_name,
            }
            error.data = undefined_resource_data
            errors.append(error)

    collector = _AnalysisKeywordsCollector(
        on_unresolved_library,
        on_unresolved_resource,
        on_resolved_library,
    )
    collect_keywords(initial_completion_context, collector)

    ast = initial_completion_context.get_ast()
    for keyword_usage_info in ast_utils.iter_keyword_usage_tokens(
        ast, collect_args_as_keywords=True
    ):
        initial_completion_context.check_cancelled()
        if contains_variable_text(keyword_usage_info.name):
            continue
        normalized_name = normalize_robot_name(keyword_usage_info.name)
        keywords_found = collector.get_keywords(normalized_name)
        if not keywords_found and keyword_usage_info.prefix:
            keywords_found = collector.get_keywords(
                normalize_robot_name(keyword_usage_info.prefix + normalized_name)
            )

        try:
            if not keywords_found:
                from robotframework_ls.impl.robot_lsp_constants import (
                    OPTION_ROBOT_LINT_UNDEFINED_KEYWORDS,
                )

                if config is not None and not config.get_setting(
                    OPTION_ROBOT_LINT_UNDEFINED_KEYWORDS, bool, True
                ):
                    continue

                node = keyword_usage_info.node
                error = create_error_from_node(
                    node,
                    "Undefined keyword: %s." % (keyword_usage_info.name,),
                    tokens=[keyword_usage_info.token],
                )
                undefined_keyword_data: ICustomDiagnosticDataUndefinedKeywordTypedDict = {
                    "kind": "undefined_keyword",
                    "name": keyword_usage_info.name,
                }
                error.data = undefined_keyword_data
                errors.append(error)

            else:
                new_keywords_found: List[IKeywordFound] = []
                if len(keywords_found) > 1:
                    # We still can't be sure, it's possible that we found the
                    # same keyword multiple times. Let's check where they're found.
                    node = keyword_usage_info.node
                    found_in = set()
                    count_not_in_stdlib = 0
                    found_in_current = False

                    for keyword_found in keywords_found:
                        if (
                            keyword_found.source
                            == initial_completion_context.original_doc.path
                        ):
                            found_in_current = True
                            for keyword_found in keywords_found:
                                # If it's defined in the current file,
                                # it overrides any other scope.
                                new_keywords_found = [
                                    k
                                    for k in keywords_found
                                    if k.source
                                    == initial_completion_context.original_doc.path
                                ]
                                for k in new_keywords_found:
                                    library_name = k.library_name
                                    library_alias = k.library_alias
                                    if library_alias:
                                        found_in.add(library_alias)
                                    else:
                                        if library_name:
                                            found_in.add(library_name)
                                        else:
                                            resource_name = keyword_found.resource_name
                                            found_in.add(resource_name)
                                count_not_in_stdlib = len(new_keywords_found)
                            break

                        library_name = keyword_found.library_name
                        if (
                            not library_name
                            or library_name.lower() not in STDLIBS_LOWER
                        ):
                            count_not_in_stdlib += 1
                            # A builtin lib is always overridden by any other place,
                            # so, don't add it to the new_keywords_found unless
                            # it was found in a non-stdlib library/resource.
                            new_keywords_found.append(keyword_found)

                        library_alias = keyword_found.library_alias
                        if library_alias:
                            found_in.add(library_alias)
                        else:
                            if library_name:
                                found_in.add(library_name)
                            else:
                                resource_name = keyword_found.resource_name
                                found_in.add(resource_name)

                    if count_not_in_stdlib > 1:
                        if found_in_current:
                            msg = f"Multiple keywords matching: '{keyword_usage_info.name}' in current file."
                        else:

                            found_in_str = "'" + "', '".join(sorted(found_in)) + "'"
                            if len(found_in) == 1:
                                msg = f"Multiple keywords matching: '{keyword_usage_info.name}' in {found_in_str}."
                            else:
                                if "." in keyword_usage_info.name:
                                    msg = f"Multiple keywords matching: '{keyword_usage_info.name}' in {found_in_str}."
                                else:
                                    msg = (
                                        f"Multiple keywords matching: '{keyword_usage_info.name}' in {found_in_str}.\n"
                                        f"Please provide the name with the full qualifier (i.e.: "
                                        f"'{sorted(found_in)[0] + '.' + keyword_usage_info.name}')."
                                    )
                        from robotframework_ls.impl.robot_lsp_constants import (
                            OPTION_ROBOT_LINT_KEYWORD_RESOLVES_TO_MULTIPLE_KEYWORDS,
                        )

                        if config is None or config.get_setting(
                            OPTION_ROBOT_LINT_KEYWORD_RESOLVES_TO_MULTIPLE_KEYWORDS,
                            bool,
                            True,
                        ):
                            error = create_error_from_node(
                                node,
                                msg,
                                tokens=[keyword_usage_info.token],
                            )
                            errors.append(error)

                from robotframework_ls.impl.robot_lsp_constants import (
                    OPTION_ROBOT_LINT_KEYWORD_CALL_ARGUMENTS,
                )

                if config is not None and not config.get_setting(
                    OPTION_ROBOT_LINT_KEYWORD_CALL_ARGUMENTS, bool, True
                ):
                    continue

                from robotframework_ls.impl.keyword_argument_analysis import (
                    KeywordArgumentAnalysis,
                )

                if new_keywords_found:
                    keywords_found = new_keywords_found

                # Still do the keyword analysis even if multiple keywords match
                # See: https://github.com/robocorp/robotframework-lsp/issues/724
                found_error_in_arg_analysis = False
                for keyword_found in keywords_found:
                    if found_error_in_arg_analysis:
                        break

                    keyword_token = keyword_usage_info.node.get_token(Token.KEYWORD)
                    if keyword_token is not None:
                        # Ok, we found the keyword, let's check if the arguments are correct.
                        keyword_argument_analysis = KeywordArgumentAnalysis(
                            keyword_found.keyword_args, keyword_found
                        )

                        for (
                            error
                        ) in keyword_argument_analysis.collect_keyword_usage_errors(
                            UsageInfoForKeywordArgumentAnalysis(
                                keyword_usage_info.node,
                                keyword_token,
                            )
                        ):
                            errors.append(error)
                            found_error_in_arg_analysis = True
                    else:
                        # Not a keyword usage, check for other cases (template/fixtures).
                        if keyword_usage_info.node.type in (
                            Token.TEMPLATE,
                            Token.TEST_TEMPLATE,
                        ):
                            # For templates the arguments are actually gotten from the test.
                            stack = keyword_usage_info.stack
                            if keyword_usage_info.node.type == Token.TEST_TEMPLATE:
                                stack = [ast]
                            for (
                                template_arguments_node_info
                            ) in ast_utils.iter_arguments_from_template(
                                stack, keyword_usage_info.node
                            ):
                                keyword_argument_analysis = KeywordArgumentAnalysis(
                                    keyword_found.keyword_args, keyword_found
                                )
                                args_tokens = template_arguments_node_info.node.tokens
                                for (
                                    error
                                ) in keyword_argument_analysis.collect_keyword_usage_errors(
                                    UsageInfoForKeywordArgumentAnalysis(
                                        template_arguments_node_info.node,
                                        args_tokens[-1],
                                        args_tokens,
                                    )
                                ):
                                    errors.append(error)
                                    found_error_in_arg_analysis = True

                for keyword_found in keywords_found:
                    if keyword_found.is_deprecated():
                        error = create_error_from_node(
                            keyword_usage_info.node,
                            f"Keyword: {keyword_usage_info.name} is deprecated",
                            tokens=[keyword_usage_info.token],
                        )
                        error.severity = DiagnosticSeverity.Hint
                        error.tags = [DiagnosticTag.Deprecated]
                        errors.append(error)
                        break

            if len(errors) >= MAX_ERRORS:
                # i.e.: Collect at most 100 errors
                break
        except:
            log.exception("Exception collecting errors")

    if len(errors) >= MAX_ERRORS:
        return errors

    for error in _collect_undefined_variables_errors(initial_completion_context):
        errors.append(error)
        if len(errors) >= MAX_ERRORS:
            # i.e.: Collect at most 100 errors
            break

    if len(errors) >= MAX_ERRORS:
        return errors

    _collect_unused_keyword_errors(initial_completion_context, errors)

    return errors


class _NoReferencesErrorsKeywordsCollector(AbstractKeywordCollector):
    def __init__(self, errors):
        from robocorp_ls_core.lsp import Error

        self.errors: List[Error] = errors

    def accepts(self, keyword_name: str) -> bool:
        return True

    def on_keyword(self, keyword_found: IKeywordFound):
        from robocorp_ls_core.lsp import Error

        if len(self.errors) >= MAX_ERRORS:
            # i.e.: Collect at most 100 errors
            return

        from robotframework_ls.impl.references import references_for_keyword_found

        completion_context = keyword_found.completion_context
        assert completion_context
        references = references_for_keyword_found(
            completion_context, keyword_found, include_declaration=False
        )
        if not references:
            start = (keyword_found.lineno, keyword_found.col_offset)
            end = (keyword_found.end_lineno, keyword_found.end_col_offset)

            error = Error(
                f"The keyword: '{keyword_found.keyword_name}' is not used in the workspace.",
                start,
                end,
                DiagnosticSeverity.Warning,
            )
            self.errors.append(error)


def _collect_unused_keyword_errors(completion_context: ICompletionContext, errors):
    from robotframework_ls.impl.robot_lsp_constants import (
        OPTION_ROBOT_LINT_UNUSED_KEYWORD,
    )
    from robotframework_ls.impl.robot_workspace import RobotWorkspace

    config = completion_context.config

    if config is not None and config.get_setting(
        OPTION_ROBOT_LINT_UNUSED_KEYWORD,
        bool,
        False,
    ):

        from robotframework_ls.impl.collect_keywords import collect_keywords_from_ast

        # The lint process usually does not have workspace indexing turned on,
        # but it's needed for finding unused references, so, turn it on now.

        workspace: Optional[RobotWorkspace] = typing.cast(
            Optional[RobotWorkspace], completion_context.workspace
        )
        if not workspace:
            log.critical("Not analyzing unused keywords because workspace is None.")
            return

        workspace_indexer = workspace.workspace_indexer
        if workspace_indexer is None:
            workspace.setup_workspace_indexer()
            workspace_indexer = workspace.workspace_indexer
            assert workspace_indexer is not None

        ast = completion_context.get_ast()
        collector = _NoReferencesErrorsKeywordsCollector(errors)
        collect_keywords_from_ast(ast, completion_context, collector)


@lru_cache(maxsize=1000)
def _skip_variable_analysis(normalized_variable_name):
    from robotframework_ls.impl.variable_resolve import is_number_var
    from robotframework_ls.impl.variable_resolve import is_python_eval_var

    if is_number_var(normalized_variable_name):
        return True

    if is_python_eval_var(normalized_variable_name):
        return True

    return False


def _env_vars_upper():
    import os

    return tuple(x.upper() for x in os.environ)


def _collect_undefined_variables_errors(initial_completion_context):
    from robotframework_ls.impl import ast_utils
    from robotframework_ls.impl.variable_resolve import normalize_variable_name
    from robotframework_ls.impl.ast_utils import create_error_from_node
    from robotframework_ls.impl.variable_resolve import robot_search_variable

    config = initial_completion_context.config
    if config is not None and not config.get_setting(
        OPTION_ROBOT_LINT_VARIABLES, bool, True
    ):
        return

    unresolved_variable_import_errors = []

    def on_unresolved_variable_import(
        completion_context: ICompletionContext,
        variable_import_name: str,
        lineno: int,
        end_lineno: int,
        col_offset: int,
        end_col_offset: int,
        error_msg: Optional[str],
        resolved_name: str,
    ):
        from robotframework_ls.impl.robot_lsp_constants import (
            OPTION_ROBOT_LINT_UNDEFINED_VARIABLE_IMPORTS,
        )

        if config is not None and not config.get_setting(
            OPTION_ROBOT_LINT_UNDEFINED_VARIABLE_IMPORTS, bool, True
        ):
            return

        doc = completion_context.doc
        if doc and doc.uri == initial_completion_context.doc.uri:
            start = (lineno - 1, col_offset)
            end = (end_lineno - 1, end_col_offset)
            if not error_msg:
                error_msg = f"Unresolved variable import: {variable_import_name}"
                if "{" in variable_import_name and resolved_name:
                    error_msg += f"\nNote: resolved name: {resolved_name}"

            error = ast_utils.Error(error_msg, start, end)
            undefined_var_import_data: ICustomDiagnosticDataUndefinedVarImportTypedDict = {
                "kind": "undefined_var_import",
                "name": variable_import_name,
                "resolved_name": resolved_name,
            }
            error.data = undefined_var_import_data
            unresolved_variable_import_errors.append(error)

    from robotframework_ls.impl import variable_completions

    ignore_variables = set()
    if config is not None:
        ignore_variables.update(
            normalize_variable_name(str(x))
            for x in config.get_setting(OPTION_ROBOT_LINT_IGNORE_VARIABLES, list, [])
        )

    ignore_environment_variables = set()
    if config is not None:
        ignore_environment_variables.update(
            str(x).upper()
            for x in config.get_setting(
                OPTION_ROBOT_LINT_IGNORE_ENVIRONMENT_VARIABLES, list, []
            )
        )

    ast = initial_completion_context.get_ast()

    globals_collector = _VariablesCollector(
        on_unresolved_variable_import=on_unresolved_variable_import
    )

    # Collect undefined variables
    variable_completions.collect_global_variables(
        initial_completion_context, globals_collector, only_current_doc=False
    )

    yield from iter(unresolved_variable_import_errors)

    env_vars_upper = None

    for token_info in ast_utils.iter_variable_references(ast):

        initial_completion_context.check_cancelled()

        if token_info.node.__class__.__name__ in (
            "ResourceImport",
            "LibraryImport",
            "VariableImport",
        ):
            # These ones are handled differently as it ends up in an unresolved
            # import.
            continue

        if (
            token_info.node.__class__.__name__ == "KeywordCall"
            and token_info.node.keyword == "Comment"
        ):
            # Special handling for 'Comment' keyword (variables are not
            # resolved when calling the 'Comment' keyword).
            # https://github.com/robocorp/robotframework-lsp/issues/665
            continue

        var_name = token_info.token.value
        var_line = token_info.token.lineno - 1  # We want it 0-based
        var_col_offset = token_info.token.col_offset

        if token_info.var_info.var_identifier == "%":

            if "=" in var_name + token_info.var_info.extended_part:
                # Consider case: %{SOME_VAR=}
                # Consider case: %{SOME_VAR=default val}
                continue

            var_name_upper = var_name.upper()
            if var_name_upper in ignore_environment_variables:
                continue

            if env_vars_upper is None:
                env_vars_upper = _env_vars_upper()

            if (
                var_name_upper not in env_vars_upper
                and not globals_collector.contains_env_variable(var_name_upper)
            ):
                # Environment variable
                yield create_error_from_node(
                    token_info.node,
                    f"Undefined environment variable: {token_info.token.value}",
                    tokens=[token_info.token],
                )
            continue

        check_names = [normalize_variable_name(var_name)]
        if token_info.var_info.extended_part.strip():

            robot_match_in_ext = robot_search_variable(
                token_info.var_info.extended_part
            )
            if robot_match_in_ext is not None and robot_match_in_ext.base:
                continue

            check_names.append(
                normalize_variable_name(var_name + token_info.var_info.extended_part)
            )

        locals_collector = None

        found = False
        for normalized_variable_name in check_names:
            if normalized_variable_name in ignore_variables:
                found = True
                break

            if _skip_variable_analysis(normalized_variable_name):
                found = True
                break

            if globals_collector.contains_variable(
                normalized_variable_name, sys.maxsize, 0
            ):
                found = True
                break

            if locals_collector is None:
                locals_collector = _VariablesCollector(lambda *args, **kwargs: None)
                local_ctx = initial_completion_context.create_copy_with_selection(
                    line=token_info.token.lineno - 1,
                    col=token_info.token.col_offset,
                )

                variable_completions.collect_local_variables(
                    local_ctx, locals_collector, token_info
                )

            if locals_collector.contains_variable(
                normalized_variable_name, var_line, var_col_offset
            ):
                found = True
                break

        if not found:
            yield create_error_from_node(
                token_info.node,
                f"Undefined variable: {token_info.token.value}",
                tokens=[token_info.token],
            )



from typing import Optional, Set, List, Dict, Iterator, Sequence
import weakref

from robocorp_ls_core.protocols import ITestInfoFromSymbolsCacheTypedDict
from robotframework_ls.impl.protocols import (
    ILibraryDoc,
    IRobotDocument,
    ISymbolsJsonListEntry,
    ICompletionContext,
    ISymbolKeywordInfo,
)
import typing
import threading
from robocorp_ls_core.robotframework_log import get_logger

log = get_logger(__name__)


class BaseSymbolsCache:
    _library_info: "Optional[weakref.ReferenceType[ILibraryDoc]]"
    _doc: "Optional[weakref.ReferenceType[IRobotDocument]]"

    def __init__(
        self,
        json_list: List[ISymbolsJsonListEntry],
        library_info: Optional[ILibraryDoc],
        doc: Optional[IRobotDocument],
        keywords_used: Set[str],
        uri: Optional[str],  # Always available if generated from doc.
        test_info: Optional[List[ITestInfoFromSymbolsCacheTypedDict]],
        global_variables_defined: Optional[Set[str]] = None,
        variable_references: Optional[Set[str]] = None,
    ):
        from robocorp_ls_core.cache import LRUCache

        self._uri = uri
        if library_info is not None:
            self._library_info = weakref.ref(library_info)
        else:
            self._library_info = None

        if doc is not None:
            self._doc = weakref.ref(doc)
        else:
            self._doc = None

        self._json_list = json_list
        self._keywords_used = keywords_used
        self._check_name_with_vars_cache_usage: LRUCache[str, bool] = LRUCache()
        self._check_name_with_vars_cache_usage_lock = threading.Lock()

        if global_variables_defined is None:
            global_variables_defined = set()
        self._global_variables_defined: Set[str] = global_variables_defined

        if variable_references is None:
            variable_references = set()
        self._variable_references: Set[str] = variable_references

        self._test_info = test_info

    def get_test_info(self) -> Optional[List[ITestInfoFromSymbolsCacheTypedDict]]:
        return self._test_info

    def get_uri(self) -> Optional[str]:
        return self._uri

    def has_keyword_usage(self, normalized_keyword_name: str) -> bool:
        ret = normalized_keyword_name in self._keywords_used
        if ret or "{" not in normalized_keyword_name:
            return ret

        with self._check_name_with_vars_cache_usage_lock:
            # The LRU is not thread safe, so, we need a lock (not ideal though as
            # it's slow)... using lru_cache would be thread safe, but we don't want to put
            # 'self' in it (so, we'd need some tricks to use it).
            # For now just use a lock for simplicity.
            found_in_cache = self._check_name_with_vars_cache_usage.get(
                normalized_keyword_name, None
            )
            if found_in_cache is not None:
                return found_in_cache

            from robotframework_ls.impl.text_utilities import (
                matches_name_with_variables,
            )

            for keyword_name_used in self._keywords_used:
                if matches_name_with_variables(
                    keyword_name_used, normalized_keyword_name
                ):
                    ret = True
                    break

            self._check_name_with_vars_cache_usage[normalized_keyword_name] = ret
        return ret

    def has_global_variable_definition(self, normalized_variable_name: str) -> bool:
        return normalized_variable_name in self._global_variables_defined

    def has_variable_reference(self, normalized_variable_name: str) -> bool:
        return normalized_variable_name in self._variable_references

    def get_json_list(self) -> List[ISymbolsJsonListEntry]:
        return self._json_list

    def get_library_info(self) -> Optional[ILibraryDoc]:
        w = self._library_info
        if w is None:
            return None
        return w()

    def get_doc(self) -> Optional[IRobotDocument]:
        w = self._doc
        if w is None:
            return None
        return w()

    def iter_keyword_info(self) -> Iterator[ISymbolKeywordInfo]:
        raise NotImplementedError("iter_keyword_info abstract in: %s", self.__class__)


class SymbolsCacheReverseIndex:
    def __init__(self):
        self._global_var_to_uris: Dict[str, Set[str]] = {}

        self._lock = threading.Lock()
        self._reindex_count = 0

        self._uris_changed = set()
        self._force_reindex = True

    def request_full_reindex(self):
        with self._lock:
            self._force_reindex = True
            self._uris_changed.clear()

    def notify_uri_changed(self, uri: str) -> None:
        with self._lock:
            if not self._force_reindex:
                self._uris_changed.add(uri)
                if len(self._uris_changed) > 1:
                    self._force_reindex = True
                    self._uris_changed.clear()

    def has_global_variable(self, normalized_var_name: str) -> bool:
        return normalized_var_name in self._global_var_to_uris

    def get_global_variable_uri_definitions(
        self, normalized_var_name: str
    ) -> Optional[Set[str]]:
        return self._global_var_to_uris.get(normalized_var_name)

    def synchronize(self, context: ICompletionContext):
        with self._lock:
            if not self._force_reindex:
                if not self._uris_changed or self._uris_changed == {context.doc.uri}:
                    # If the only thing changed is the current uri (or if there
                    # were no changes), don't do a workspace-wide update.
                    return

            # Reset synchronize-related flags.
            self._force_reindex = False
            self._uris_changed.clear()

            self._reindex_count += 1
            self._compute_new_symbols_cache_reverse_index_state(context)

    def dispose(self):
        self._global_var_to_uris = {}

    def _compute_new_symbols_cache_reverse_index_state(
        self, context: ICompletionContext
    ) -> None:
        from robotframework_ls.impl.workspace_symbols import iter_symbols_caches

        new_global_var_to_uris: Dict[str, Set[str]] = {}
        symbols_cache: BaseSymbolsCache

        # Note: always update as a whole.
        it = typing.cast(Iterator[BaseSymbolsCache], iter_symbols_caches("", context))

        try:
            for symbols_cache in it:
                uri = symbols_cache.get_uri()
                if not uri:
                    continue

                for global_var_name in symbols_cache._global_variables_defined:
                    s = new_global_var_to_uris.get(global_var_name)
                    if s is None:
                        s = new_global_var_to_uris[global_var_name] = set()
                    s.add(uri)
        except:
            log.exception("Exception computing symbols cache reverse index.")
            raise  # Maybe it was cancelled (or we had another error).
        else:
            # ok, it worked, let's actually update our internal state.
            self._global_var_to_uris = new_global_var_to_uris



# Note: there's a copy of these settings at:
# robotframework_interactive.server.rf_interpreter_ls_config
# which should be kept up-to-date with this one.

from typing import List
from .robot_generated_lsp_constants import *  # @UnusedWildImport

# Some aliases for backward compatibility.
OPTION_ROBOT_COMPLETION_SECTION_HEADERS_FORM = (
    OPTION_ROBOT_COMPLETIONS_SECTION_HEADERS_FORM
)
OPTION_ROBOT_COMPLETION_KEYWORDS_ARGUMENTS_SEPARATOR = (
    OPTION_ROBOT_COMPLETIONS_KEYWORDS_ARGUMENTS_SEPARATOR
)
OPTION_ROBOT_SHOW_CODE_LENSES = OPTION_ROBOT_CODE_LENS_ENABLE
OPTION_ROBOT_COMPLETION_KEYWORDS_FORMAT = OPTION_ROBOT_COMPLETIONS_KEYWORDS_FORMAT
OPTION_ROBOT_VARIABLES_LOAD_FROM_ARGUMENTS_FILE = (
    OPTION_ROBOT_LOAD_VARIABLES_FROM_ARGUMENTS_FILE
)


OPTION_ROBOT_CODE_FORMATTER_ROBOTIDY = "robotidy"
OPTION_ROBOT_CODE_FORMATTER_BUILTIN_TIDY = "builtinTidy"

OPTION_ROBOT_FLOW_EXPLORER_THEME_LIGHT = "light"
OPTION_ROBOT_FLOW_EXPLORER_THEME_DARK = "dark"

OPTION_ROBOT_COMPLETION_KEYWORDS_FORMAT_FIRST_UPPER = "First upper"
OPTION_ROBOT_COMPLETION_KEYWORDS_FORMAT_TITLE = "Title Case"
OPTION_ROBOT_COMPLETION_KEYWORDS_FORMAT_ALL_LOWER = "all lower"
OPTION_ROBOT_COMPLETION_KEYWORDS_FORMAT_ALL_UPPER = "ALL UPPER"

OPTION_ROBOT_COMPLETION_SECTION_HEADERS_FORM_PLURAL = "plural"
OPTION_ROBOT_COMPLETION_SECTION_HEADERS_FORM_SINGULAR = "singular"
OPTION_ROBOT_COMPLETION_SECTION_HEADERS_FORM_BOTH = "both"


# Options which must be set as environment variables.
ENV_OPTION_ROBOT_DAP_TIMEOUT = "ROBOT_DAP_TIMEOUT"


CHECK_IF_LIBRARIES_INSTALLED: List[str] = [
    # robotframework-appiumlibrary
    "AppiumLibrary",
    # robotframework-datadriver
    "DataDriver",
    # robotframework-mainframe3270
    "Mainframe3270",
    # robotframework-sshlibrary
    "SSHLibrary",
    # robotframework-whitelibrary
    "WhiteLibrary",
    # rpaframework-google
    "RPA.Cloud.Google",
    # rpaframework
    "RPA.Archive",
    "RPA.Browser",
    "RPA.Browser.Selenium",
    "RPA.Cloud.AWS",
    "RPA.Cloud.Azure",
    "RPA.Crypto",
    "RPA.Database",
    "RPA.Desktop",
    "RPA.Desktop.Clipboard",
    "RPA.Desktop.OperatingSystem",
    "RPA.Desktop.Windows",
    "RPA.Dialogs",
    "RPA.Email.Exchange",
    "RPA.Email.ImapSmtp",
    "RPA.Excel.Application",
    "RPA.Excel.Files",
    "RPA.FileSystem",
    "RPA.FTP",
    "RPA.HTTP",
    "RPA.Images",
    "RPA.JavaAccessBridge",
    "RPA.JSON",
    "RPA.Netsuite",
    "RPA.Notifier",
    "RPA.Outlook.Application",
    "RPA.PDF",
    "RPA.Robocloud.Items",
    "RPA.Robocloud.Secrets",
    "RPA.Robocorp.Process",
    "RPA.Robocorp.Vault",
    "RPA.Robocorp.WorkItems",
    "RPA.RobotLogListener",
    "RPA.Salesforce",
    "RPA.SAP",
    "RPA.Slack",
    "RPA.Tables",
    "RPA.Tasks",
    "RPA.Twitter",
    "RPA.Word.Application",
]



import os
import re
import sys
import textwrap
from pathlib import Path
from typing import Any, Dict, List, Optional, Pattern, Tuple, Union
from robotidy import skip


CLI_OPTIONS_LIST = [
    {
        "name": "Run only selected transformers",
        "options": ["--transform"],
    },
    {
        "name": "Work modes",
        "options": ["--overwrite", "--diff", "--check", "--force-order"],
    },
    {
        "name": "Documentation",
        "options": ["--list", "--desc"],
    },
    {
        "name": "Configuration",
        "options": ["--configure", "--config"],
    },
    {
        "name": "Global formatting settings",
        "options": [
            "--spacecount",
            "--indent",
            "--continuation-indent",
            "--line-length",
            "--lineseparator",
            "--separator",
            "--startline",
            "--endline",
        ],
    },
    {
        "name": "File exclusion",
        "options": ["--exclude", "--extend-exclude", "--skip-gitignore"],
    },
    skip.option_group,
    {
        "name": "Other",
        "options": [
            "--target-version",
            "--verbose",
            "--color",
            "--output",
            "--version",
            "--help",
        ],
    },
]

try:
    import rich_click as click

    click.rich_click.USE_RICH_MARKUP = True
    click.rich_click.USE_MARKDOWN = True
    click.rich_click.STYLE_OPTION = "bold sky_blue3"
    click.rich_click.STYLE_SWITCH = "bold sky_blue3"
    click.rich_click.STYLE_METAVAR = "bold white"
    click.rich_click.STYLE_OPTION_DEFAULT = "grey37"
    click.rich_click.STYLE_OPTIONS_PANEL_BORDER = "grey66"
    click.rich_click.STYLE_USAGE = "magenta"
    click.rich_click.OPTION_GROUPS = {
        "robotidy": CLI_OPTIONS_LIST,
        "python -m robotidy": CLI_OPTIONS_LIST,
    }

except ImportError:
    import click

from robotidy.app import Robotidy
from robotidy.config import Config, FormattingConfig
from robotidy.decorators import catch_exceptions
from robotidy.files import DEFAULT_EXCLUDES, find_and_read_config, read_pyproject_config

try:
    from robotidy.rich_console import console
except:

    class console:
        def print(self, *args, **kwargs):
            print(*args, **kwargs)

    console = console()
from robotidy.transformers import load_transformers
from robotidy.utils import (
    ROBOT_VERSION,
    RecommendationFinder,
    TargetVersion,
    split_args_from_name_or_path,
)
from robotidy.version import __version__


CONTEXT_SETTINGS = dict(help_option_names=["-h", "--help"])


class TransformType(click.ParamType):
    name = "transform"

    def convert(self, value, param, ctx):
        name, args = split_args_from_name_or_path(value.replace(" ", ""))
        return name, args


def parse_opt(opt):
    while opt and opt[0] == "-":
        opt = opt[1:]
    return opt.replace("-", "_")


def validate_config_options(params, config):
    if params is None:
        return
    allowed = {parse_opt(opt) for param in params for opt in param.opts}
    for conf in config:
        if conf not in allowed:
            rec_finder = RecommendationFinder()
            similar = rec_finder.find(conf, list(allowed))
            raise click.NoSuchOption(conf, possibilities=similar)


def read_config(
    ctx: click.Context, param: click.Parameter, value: Optional[str]
) -> Optional[str]:
    # if --config was not used, try to find pyproject.toml or robotidy.toml file
    if value:
        config = read_pyproject_config(value)
    else:
        config = find_and_read_config(ctx.params["src"] or (str(Path(".").resolve()),))
    if not config:
        return
    # Sanitize the values to be Click friendly. For more information please see:
    # https://github.com/psf/black/issues/1458
    # https://github.com/pallets/click/issues/1567
    config = {
        k: str(v) if not isinstance(v, (list, dict)) else v for k, v in config.items()
    }
    if "src" in config:
        config["src"] = tuple(config["src"])
    validate_config_options(ctx.command.params, config)
    default_map: Dict[str, Any] = {}
    if ctx.default_map:
        default_map.update(ctx.default_map)
    default_map.update(config)
    ctx.default_map = default_map


def validate_regex_callback(
    ctx: click.Context,
    param: click.Parameter,
    value: Optional[str],
) -> Optional[Pattern]:
    return validate_regex(value)


def validate_target_version(
    ctx: click.Context,
    param: Union[click.Option, click.Parameter],
    value: Optional[str],
) -> Optional[int]:
    if value is None:
        return ROBOT_VERSION.major
    version = TargetVersion[value.upper()].value
    if version > ROBOT_VERSION.major:
        raise click.BadParameter(
            f"Target Robot Framework version ({version}) should not be higher than installed version ({ROBOT_VERSION})."
        )
    return version


def validate_regex(value: Optional[str]) -> Optional[Pattern]:
    try:
        return re.compile(value) if value is not None else None
    except re.error:
        raise click.BadParameter("Not a valid regular expression")


def print_transformer_docs(transformer):
    from rich.markdown import Markdown

    md = Markdown(
        str(transformer), code_theme="native", inline_code_lexer="robotframework"
    )
    console.print(md)


def print_description(name: str, target_version: int):
    transformers = load_transformers(
        None, {}, allow_disabled=True, target_version=target_version
    )
    transformer_by_names = {
        transformer.name: transformer for transformer in transformers
    }
    if name == "all":
        for transformer in transformers:
            print_transformer_docs(transformer)
    elif name in transformer_by_names:
        print_transformer_docs(transformer_by_names[name])
    else:
        rec_finder = RecommendationFinder()
        similar = rec_finder.find_similar(name, transformer_by_names.keys())
        click.echo(f"Transformer with the name '{name}' does not exist.{similar}")
        return 1
    return 0


def print_transformers_list(target_version: int):
    from rich.table import Table

    table = Table(title="Transformers", header_style="bold red")
    table.add_column("Name", justify="left", no_wrap=True)
    table.add_column("Enabled by default")
    transformers = load_transformers(
        None, {}, allow_disabled=True, target_version=target_version
    )
    for transformer in transformers:
        decorated_enable = (
            "Yes" if transformer.enabled_by_default else "[bold magenta]No"
        )
        table.add_row(transformer.name, decorated_enable)
    console.print(table)
    console.print("Transformers are listed in the order they are run by default.")
    console.print(
        "To see detailed docs run:\n"
        "    [bold]robotidy --desc [bold magenta]transformer_name[/][/]\n"
        "or\n"
        "    [bold]robotidy --desc [bold blue]all[/][/]\n\n"
        "Non-default transformers needs to be selected explicitly with [bold cyan]--transform[/] or "
        "configured with param `enabled=True`.\n"
    )


@click.command(context_settings=CONTEXT_SETTINGS)
@click.option(
    "--transform",
    "-t",
    type=TransformType(),
    multiple=True,
    metavar="TRANSFORMER_NAME",
    help="Transform files from \[PATH(S)] with given transformer",
)
@click.option(
    "--configure",
    "-c",
    type=TransformType(),
    multiple=True,
    metavar="TRANSFORMER_NAME:PARAM=VALUE",
    help="Configure transformers",
)
@click.argument(
    "src",
    nargs=-1,
    type=click.Path(
        exists=True, file_okay=True, dir_okay=True, readable=True, allow_dash=True
    ),
    is_eager=True,
    metavar="[PATH(S)]",
)
@click.option(
    "--exclude",
    type=str,
    callback=validate_regex_callback,
    help=(
        "A regular expression that matches files and directories that should be"
        " excluded on recursive searches. An empty value means no paths are excluded."
        " Use forward slashes for directories on all platforms."
    ),
    show_default=f"{DEFAULT_EXCLUDES}",
)
@click.option(
    "--extend-exclude",
    type=str,
    callback=validate_regex_callback,
    help=(
        "Like **--exclude**, but adds additional files and directories on top of the"
        " excluded ones. (Useful if you simply want to add to the default)"
    ),
)
@click.option(
    "--skip-gitignore",
    is_flag=True,
    show_default=True,
    help="Skip **.gitignore** files and do not ignore files listed inside.",
)
@click.option(
    "--config",
    type=click.Path(
        exists=True,
        file_okay=True,
        dir_okay=False,
        readable=True,
        allow_dash=False,
        path_type=str,
    ),
    is_eager=True,
    callback=read_config,
    help="Read configuration from FILE path.",
)
@click.option(
    "--overwrite/--no-overwrite",
    default=None,
    help="Write changes back to file",
    show_default=True,
)
@click.option(
    "--diff",
    is_flag=True,
    help="Output diff of each processed file.",
    show_default=True,
)
@click.option(
    "--color/--no-color",
    default=True,
    help="Enable ANSI coloring the output",
    show_default=True,
)
@click.option(
    "--check",
    is_flag=True,
    help="Don't overwrite files and just return status. Return code 0 means nothing would change. "
    "Return code 1 means that at least 1 file would change. Any internal error will overwrite this status.",
    show_default=True,
)
@click.option(
    "-s",
    "--spacecount",
    type=click.types.INT,
    default=4,
    help="The number of spaces between cells",
    show_default=True,
)
@click.option(
    "--indent",
    type=click.types.INT,
    default=None,
    help="The number of spaces to be used as indentation",
    show_default="same as --spacecount value",
)
@click.option(
    "--continuation-indent",
    type=click.types.INT,
    default=None,
    help="The number of spaces to be used as separator after ... (line continuation) token",
    show_default="same as --spacecount value]",
)
@click.option(
    "-ls",
    "--lineseparator",
    type=click.types.Choice(["native", "windows", "unix", "auto"]),
    default="native",
    help="""
    Line separator to use in the outputs:
    - **native**:  use operating system's native line endings
    - windows: use Windows line endings (CRLF)
    - unix:    use Unix line endings (LF)
    - auto:    maintain existing line endings (uses what's used in the first line)
    """,
    show_default=True,
)
@click.option(
    "--separator",
    type=click.types.Choice(["space", "tab"]),
    default="space",
    help="""
    Token separator to use in the outputs:
    - **space**:   use --spacecount spaces to separate tokens
    - tab:     use a single tabulation to separate tokens
    """,
    show_default=True,
)
@click.option(
    "-sl",
    "--startline",
    default=None,
    type=int,
    help="Limit robotidy only to selected area. If **--endline** is not provided, format text only at **--startline**. "
    "Line numbers start from 1.",
)
@click.option(
    "-el",
    "--endline",
    default=None,
    type=int,
    help="Limit robotidy only to selected area. Line numbers start from 1.",
)
@click.option(
    "--line-length",
    default=120,
    type=int,
    help="Max allowed characters per line",
    show_default=True,
)
@click.option(
    "--list",
    "-l",
    is_eager=True,
    is_flag=True,
    help="List available transformers and exit.",
)
@click.option(
    "--desc",
    "-d",
    default=None,
    metavar="TRANSFORMER_NAME",
    help="Show documentation for selected transformer.",
)
@click.option(
    "--output",
    "-o",
    type=click.Path(file_okay=True, dir_okay=False, writable=True, allow_dash=False),
    default=None,
    metavar="PATH",
    help="Use this option to override file destination path.",
)
@click.option(
    "-v", "--verbose", is_flag=True, help="More verbose output", show_default=True
)
@click.option(
    "--force-order",
    is_flag=True,
    help="Transform files using transformers in order provided in cli",
)
@click.option(
    "--target-version",
    "-tv",
    type=click.Choice([v.name.lower() for v in TargetVersion], case_sensitive=False),
    callback=validate_target_version,
    help="Only enable transformers supported in set target version",
    show_default="installed Robot Framework version",
)
@skip.documentation_option
@skip.return_values_option
@skip.keyword_call_option
@skip.keyword_call_pattern_option
@skip.settings_option
@skip.arguments_option
@skip.setup_option
@skip.teardown_option
@skip.timeout_option
@skip.template_option
@skip.return_option
@skip.tags_option
@skip.block_comments_option
@click.version_option(version=__version__, prog_name="robotidy")
@click.pass_context
@catch_exceptions
def cli(
    ctx: click.Context,
    transform: List[Tuple[str, List]],
    configure: List[Tuple[str, List]],
    src: Tuple[str, ...],
    exclude: Optional[Pattern],
    extend_exclude: Optional[Pattern],
    skip_gitignore: bool,
    overwrite: bool,
    diff: bool,
    color: bool,
    check: bool,
    spacecount: int,
    indent: Optional[int],
    continuation_indent: Optional[int],
    lineseparator: str,
    verbose: bool,
    config: Optional[str],
    separator: Optional[str],
    startline: Optional[int],
    endline: Optional[int],
    line_length: int,
    list: bool,
    desc: Optional[str],
    output: Optional[Path],
    force_order: bool,
    target_version: int,
    skip_documentation: bool,
    skip_return_values: bool,
    skip_keyword_call: List[str],
    skip_keyword_call_pattern: List[str],
    skip_settings: bool,
    skip_arguments: bool,
    skip_setup: bool,
    skip_teardown: bool,
    skip_timeout: bool,
    skip_template: bool,
    skip_return: bool,
    skip_tags: bool,
    skip_block_comments: bool,
):
    """
    Robotidy is a tool for formatting Robot Framework source code.
    Full documentation available at <https://robotidy.readthedocs.io> .
    """
    if list:
        print_transformers_list(target_version)
        sys.exit(0)
    if desc is not None:
        return_code = print_description(desc, target_version)
        sys.exit(return_code)
    if not src:
        if ctx.default_map is not None:
            src = ctx.default_map.get("src", None)
        if not src:
            print(
                "No source path provided. Run robotidy --help to see how to use robotidy"
            )
            sys.exit(1)

    if exclude is None:
        exclude = re.compile(DEFAULT_EXCLUDES)

    if config and verbose:
        click.echo(f"Loaded {config} configuration file")

    if overwrite is None:
        # None is default, with check not set -> overwrite, with check set -> overwrite only when overwrite flag is set
        overwrite = not check

    if color:
        color = "NO_COLOR" not in os.environ

    skip_config = skip.SkipConfig(
        documentation=skip_documentation,
        return_values=skip_return_values,
        keyword_call=skip_keyword_call,
        keyword_call_pattern=skip_keyword_call_pattern,
        settings=skip_settings,
        arguments=skip_arguments,
        setup=skip_setup,
        teardown=skip_teardown,
        template=skip_template,
        timeout=skip_timeout,
        return_statement=skip_return,
        tags=skip_tags,
        block_comments=skip_block_comments,
    )

    formatting = FormattingConfig(
        space_count=spacecount,
        indent=indent,
        continuation_indent=continuation_indent,
        line_sep=lineseparator,
        start_line=startline,
        separator=separator,
        end_line=endline,
        line_length=line_length,
    )
    config = Config(
        formatting=formatting,
        skip=skip_config,
        transformers=transform,
        transformers_config=configure,
        src=src,
        exclude=exclude,
        extend_exclude=extend_exclude,
        skip_gitignore=skip_gitignore,
        overwrite=overwrite,
        show_diff=diff,
        verbose=verbose,
        check=check,
        output=output,
        force_order=force_order,
        target_version=target_version,
        color=color,
    )
    tidy = Robotidy(config=config)
    status = tidy.transform_files()
    sys.exit(status)



import sys

from _pydevd_bundle.pydevd_constants import PANDAS_MAX_ROWS, PANDAS_MAX_COLS, PANDAS_MAX_COLWIDTH
from _pydevd_bundle.pydevd_extension_api import TypeResolveProvider, StrPresentationProvider
from _pydevd_bundle.pydevd_resolver import inspect, MethodWrapperType
from _pydevd_bundle.pydevd_utils import Timer

from .pydevd_helpers import find_mod_attr


def _get_dictionary(obj, replacements):
    ret = dict()
    cls = obj.__class__
    for attr_name in dir(obj):

        # This is interesting but it actually hides too much info from the dataframe.
        # attr_type_in_cls = type(getattr(cls, attr_name, None))
        # if attr_type_in_cls == property:
        #     ret[attr_name] = '<property (not computed)>'
        #     continue

        timer = Timer()
        try:
            replacement = replacements.get(attr_name)
            if replacement is not None:
                ret[attr_name] = replacement
                continue

            attr_value = getattr(obj, attr_name, '<unable to get>')
            if inspect.isroutine(attr_value) or isinstance(attr_value, MethodWrapperType):
                continue
            ret[attr_name] = attr_value
        except Exception as e:
            ret[attr_name] = '<error getting: %s>' % (e,)
        finally:
            timer.report_if_getting_attr_slow(cls, attr_name)

    return ret


class PandasDataFrameTypeResolveProvider(object):

    def can_provide(self, type_object, type_name):
        data_frame_class = find_mod_attr('pandas.core.frame', 'DataFrame')
        return data_frame_class is not None and issubclass(type_object, data_frame_class)

    def resolve(self, obj, attribute):
        return getattr(obj, attribute)

    def get_dictionary(self, obj):
        replacements = {
            # This actually calls: DataFrame.transpose(), which can be expensive, so,
            # let's just add some string representation for it.
            'T': '<transposed dataframe -- debugger:skipped eval>',

            # This creates a whole new dict{index: Series) for each column. Doing a
            # subsequent repr() from this dict can be very slow, so, don't return it.
            '_series': '<dict[index:Series] -- debugger:skipped eval>',

            'style': '<pandas.io.formats.style.Styler -- debugger: skipped eval>',
        }
        return _get_dictionary(obj, replacements)

    def get_str(self, df):
        # The default repr depends on the settings of:
        # pandas.set_option('display.max_columns', None)
        # pandas.set_option('display.max_rows', None)
        # which can make the repr **very** slow on some cases, so, let's use a
        # version which

        return df.to_string(
            max_rows=PANDAS_MAX_ROWS,
            max_cols=PANDAS_MAX_COLS,
            max_colwidth=PANDAS_MAX_COLWIDTH,
            show_dimensions=True,
        )


class PandasSeriesTypeResolveProvider(object):

    def can_provide(self, type_object, type_name):
        series_class = find_mod_attr('pandas.core.series', 'Series')
        return series_class is not None and issubclass(type_object, series_class)

    def resolve(self, obj, attribute):
        return getattr(obj, attribute)

    def get_dictionary(self, obj):
        replacements = {
            # This actually calls: DataFrame.transpose(), which can be expensive, so,
            # let's just add some string representation for it.
            'T': '<transposed dataframe -- debugger:skipped eval>',

            # This creates a whole new dict{index: Series) for each column. Doing a
            # subsequent repr() from this dict can be very slow, so, don't return it.
            '_series': '<dict[index:Series] -- debugger:skipped eval>',

            'style': '<pandas.io.formats.style.Styler -- debugger: skipped eval>',
        }
        return _get_dictionary(obj, replacements)

    def get_str(self, series):
        return (series.to_string(
            max_rows=PANDAS_MAX_ROWS,
        ))


class PandasStylerTypeResolveProvider(object):

    def can_provide(self, type_object, type_name):
        series_class = find_mod_attr('pandas.io.formats.style', 'Styler')
        return series_class is not None and issubclass(type_object, series_class)

    def resolve(self, obj, attribute):
        return getattr(obj, attribute)

    def get_dictionary(self, obj):
        replacements = {
            'data': '<Styler data -- debugger:skipped eval>',

            '__dict__': '<dict -- debugger: skipped eval>',
        }
        return _get_dictionary(obj, replacements)


if not sys.platform.startswith("java"):
    TypeResolveProvider.register(PandasDataFrameTypeResolveProvider)
    StrPresentationProvider.register(PandasDataFrameTypeResolveProvider)

    TypeResolveProvider.register(PandasSeriesTypeResolveProvider)
    StrPresentationProvider.register(PandasSeriesTypeResolveProvider)

    TypeResolveProvider.register(PandasStylerTypeResolveProvider)



import datetime
import json
from typing import Optional, Any, Iterator
import sys
import functools
from io import StringIO

__version__ = "0.0.4"
version_info = [int(x) for x in __version__.split(".")]


_convert = {
    "gb": lambda s: s * 1e9,
    "g": lambda s: s * 1e9,
    "mb": lambda s: s * 1e6,
    "m": lambda s: s * 1e6,
    "kb": lambda s: s * 1000,
    "k": lambda s: s * 1000,
    "b": lambda s: s,
    "": lambda s: s,
}


def _convert_to_bytes(s):
    initial = s
    num = []
    while s and (s[0].isdigit() or s[0] == "."):
        num.append(s[0])
        s = s[1:]
    num = float("".join(num))
    unit = s.strip()
    conv = _convert.get(unit.lower())
    if conv is None:
        raise ValueError(f"Cannot get in bytes: {initial}")

    return conv(num)


def log_error(func):
    @functools.wraps(func)
    def new_func(self, *args, **kwargs):
        import traceback

        try:
            return func(self, *args, **kwargs)
        except Exception as e:
            s = StringIO()
            traceback.print_exc(file=s)
            self._robot_output_impl.log_message(
                "ERROR",
                f"RFStream internal error: {e}\n{s.getvalue()}",
                self._robot_output_impl.get_time_delta(),
                False,
            )

    return new_func


class RFStream:
    """
    This is the class which should be registered as a listener.
    """

    # V3 would be nicer but it doesn't support keywords...
    ROBOT_LISTENER_API_VERSION = 2

    def __init__(self, *args, **kwargs):
        from robot_out_stream._impl import _RobotOutputImpl, _Config

        # Note: expected to be used just when used in-memory.
        config = _Config(kwargs.get("__uuid__"))

        check_args = ["--dir=", "--max-file-size=", "--max-files=", "--log="]
        for arg in args:
            for check_arg in check_args:
                if arg.startswith(check_arg):
                    kwargs[check_arg[:-1]] = arg[len(check_arg) :]

        config.output_dir = kwargs.get("--dir", "./out_rfstream").replace(
            "<COLON>", ":"
        )
        if config.output_dir == "None":
            config.output_dir = None

        config.log_html = kwargs.get("--log")
        if config.log_html:
            config.log_html = config.log_html.replace("<COLON>", ":")

        max_file_size_arg = kwargs.get("--max-file-size", "1MB")
        config.max_file_size_in_bytes = _convert_to_bytes(max_file_size_arg)
        config.max_files = int(kwargs.get("--max-files", "5"))

        if config.max_file_size_in_bytes < 1000:
            raise ValueError(
                f"Cannot generate logs where the max file size in bytes is less that 1000 bytes."
                f" Found: {config.max_file_size_in_bytes}."
                f" Arg: {max_file_size_arg}."
            )

        # Attributes expected to be used just when used in-memory.
        config.write = kwargs.get("__write__")
        config.initial_time = kwargs.get("__initial_time__")
        config.additional_info = kwargs.get("__additional_info__")

        self._robot_output_impl = _RobotOutputImpl(config)

    @property
    def robot_output_impl(self):
        return self._robot_output_impl

    @property
    def initial_time(self) -> datetime.datetime:
        return self._robot_output_impl.initial_time

    def _get_time_delta(self, attributes) -> float:
        # i.e.: in general the time delta will not be there
        # it's only there for the case when we're reading
        # from the output.xml.
        time_delta = attributes.get("timedelta")
        if time_delta is not None:
            return time_delta
        return self._robot_output_impl.get_time_delta()

    @log_error
    def start_suite(self, name, attributes):
        # {
        #     "id": "s1",
        #     "longname": "Robot1",
        #     "doc": "",
        #     "metadata": {},
        #     "starttime": "20221003 14:20:02.195",
        #     "tests": ["First task", "Second task"],
        #     "suites": [],
        #     "totaltests": 2,
        #     "source": "C:\\Users\\...\\robot1.robot",
        # }
        return self._robot_output_impl.start_suite(
            name,
            attributes["id"],
            attributes["source"],
            self._get_time_delta(attributes),
        )

    @log_error
    def end_suite(self, name, attributes):
        # {
        #     "id": "s1",
        #     "longname": "Robot1",
        #     "doc": "",
        #     "metadata": {},
        #     "starttime": "20221004 09:38:40.271",
        #     "endtime": "20221004 09:38:40.323",
        #     "elapsedtime": 52,
        #     "status": "FAIL",
        #     "message": "",
        #     "tests": ["First task", "Second task"],
        #     "suites": [],
        #     "totaltests": 2,
        #     "source": "C:\\Users\\fabio\\AppData\\Local\\Temp\\pytest-of-fabio\\pytest-184\\test_robot_out_stream0\\test_robot_out_stream\\robot1.robot",
        #     "statistics": "2 tasks, 1 passed, 1 failed",
        # }
        return self._robot_output_impl.end_suite(
            attributes["id"], attributes["status"], self._get_time_delta(attributes)
        )

    @log_error
    def start_test(self, name, attributes):
        # {
        #     "id": "s1-t1",
        #     "longname": "Robot1.First task",
        #     "doc": "",
        #     "tags": [],
        #     "lineno": 11,
        #     "source": "C:\\Users\\fabio\\...\\robot1.robot",
        #     "starttime": "20221003 14:20:02.231",
        #     "template": "",
        #     "originalname": "First task",
        # }
        return self._robot_output_impl.start_test(
            name,
            attributes["id"],
            attributes.get(
                "lineno"
            ),  # The source is already given by the suite (no need to repeat)
            self._get_time_delta(attributes),
            attributes.get("tags"),
        )

    @log_error
    def send_tag(self, tag: str):
        return self._robot_output_impl.send_tag(tag)

    @log_error
    def send_info(self, info: str):
        return self._robot_output_impl.send_info(info)

    @log_error
    def send_start_time_delta(self, time_delta_in_seconds: float):
        return self._robot_output_impl.send_start_time_delta(time_delta_in_seconds)

    @log_error
    def end_test(self, name, attributes):
        # {
        #     "id": "s1-t2",
        #     "longname": "Robot1.Second task",
        #     "doc": "",
        #     "tags": [],
        #     "lineno": 15,
        #     "source": "C:\\Users\\fabio\\AppData\\Local\\Temp\\pytest-of-fabio\\pytest-187\\test_robot_out_stream0\\test_robot_out_stream\\robot1.robot",
        #     "starttime": "20221004 16:23:10.403",
        #     "endtime": "20221004 16:23:10.412",
        #     "elapsedtime": 9,
        #     "status": "FAIL",
        #     "message": "Failed execution for some reason...",
        #     "template": "",
        #     "originalname": "Second task",
        # }
        return self._robot_output_impl.end_test(
            attributes["id"],
            attributes["status"],
            attributes["message"],
            self._get_time_delta(attributes),
        )

    class _Sentinel:
        pass

    @log_error
    def start_keyword(self, name, attributes):

        # {
        #     "doc": "Does absolutely nothing.",
        #     "assign": [],
        #     "tags": [],
        #     "lineno": 7,
        #     "source": "C:\\Users\\fabio\\AppData\\Local\\Temp\\pytest-of-fabio\\pytest-170\\test_robot_out_stream0\\test_robot_out_stream\\robot1.robot",
        #     "type": "KEYWORD",
        #     "status": "NOT SET",
        #     "starttime": "20221003 16:20:21.234",
        #     "kwname": "No Operation",
        #     "libname": "BuiltIn",
        #     "args": [],
        # }
        source: Optional[str] = attributes.get("source", self._Sentinel)
        lineno: Optional[int] = attributes.get("lineno", self._Sentinel)
        if source is self._Sentinel or lineno is self._Sentinel:
            # I.e.: it was not passed at all (if it was passed and None,
            # keep it as None: xml conversion use-case).
            source = None
            lineno = -1

            # HACK for RF 3: try to get the location (since it's not available).
            f: Optional[Any]
            f = sys._getframe()
            while f is not None:
                if f.f_code.co_name == "run_step":
                    step = f.f_locals.get("step")
                    if step is not None:
                        try:
                            source = step.source
                            lineno = step.lineno
                        except AttributeError:
                            pass
                    break  # Break when run_step is found anyways.

                f = f.f_back

        return self._robot_output_impl.start_keyword(
            attributes["kwname"],
            attributes.get("libname"),
            attributes.get("type"),
            attributes.get("doc"),
            source,
            lineno,
            self._get_time_delta(attributes),
            attributes.get("args"),
            attributes.get("assign"),
        )

    @log_error
    def end_keyword(self, name, attributes):
        # {
        #     "doc": "Does absolutely nothing.",
        #     "assign": [],
        #     "tags": [],
        #     "lineno": 7,
        #     "source": "C:\\Users\\fabio\\AppData\\Local\\Temp\\pytest-of-fabio\\pytest-191\\test_robot_out_stream0\\test_robot_out_stream\\robot1.robot",
        #     "type": "KEYWORD",
        #     "status": "PASS",
        #     "starttime": "20221004 16:27:46.959",
        #     "endtime": "20221004 16:27:46.959",
        #     "elapsedtime": 0,
        #     "kwname": "No Operation",
        #     "libname": "BuiltIn",
        #     "args": [],
        # }
        name = attributes["kwname"]
        libname = attributes.get("libname")

        return self._robot_output_impl.end_keyword(
            name, libname, attributes["status"], self._get_time_delta(attributes)
        )

    @log_error
    def log_message(self, message, skip_error=True):
        # {
        #     "timestamp": "20221026 10:00:31.591",
        #     "message": "${dct} = {'a': '1', 'b': '1'}",
        #     "level": "INFO",
        #     "html": "no",
        # }
        level = message["level"]
        if level not in ("ERROR", "FAIL", "WARN", "INFO"):
            # Exclude TRACE/DEBUG/HTML for now (we could make that configurable...)
            return

        if skip_error and level in ("ERROR",):
            # We do this because in RF all the calls to 'log_message'
            # also generate a call to 'message', so, we want to skip
            # one of those (but note that the other way around isn't true
            # and some errors such as import errors are only reported
            # in 'message' and not 'log_message').
            return

        html = message.get("html")
        return self._robot_output_impl.log_message(
            level, message["message"], self._get_time_delta(message), html
        )

    @log_error
    def message(self, message):
        if message["level"] in ("FAIL", "ERROR"):
            return self.log_message(message, skip_error=False)

    @log_error
    def close(self):
        self.robot_output_impl.close()


def iter_decoded_log_format(stream) -> Iterator[dict]:
    """
    :param stream:
        The stream which should be iterated in (anything with a `readlines()` method).

    :returns:
        An iterator which will decode the messages and provides a dictionary for
        each message found.

        Example of messages provided:

        {'message_type': 'V', 'version': '1'}
        {'message_type': 'T', 'initial_time': '2022-10-31T07:45:57.116'}
        {'message_type': 'ID', 'part': 1, 'id': 'gen-from-output-xml'}
        {'message_type': 'SS', 'name': 'Robot Check', 'suite_id': 's1', 'suite_source': 'x:\\vscode-robot\\local_test\\robot_check', 'time_delta_in_seconds': 0.3}
        {'message_type': 'ST', 'name': 'My test', 'suite_id': 's1-s1-t1', 'lineno': 5, 'time_delta_in_seconds': 0.2}
    """
    from ._decoder import iter_decoded_log_format

    return iter_decoded_log_format(stream)



# fmt: off
# Warning: Don't edit file (autogenerated from python -m dev codegen).
from typing import List

ROBOT_RUN_TEST = "robot.runTest"  # Run Test/Task
ROBOT_DEBUG_TEST = "robot.debugTest"  # Debug Test/Task
ROBOT_RUN_SUITE = "robot.runSuite"  # Run Tests/Tasks Suite
ROBOT_DEBUG_SUITE = "robot.debugSuite"  # Debug Tests/Tasks Suite
ROBOT_INTERACTIVE_SHELL = "robot.interactiveShell"  # Start Interactive Console
ROBOT_INTERNAL_RFINTERACTIVE_START = "robot.internal.rfinteractive.start"  # Create Interactive Console
ROBOT_INTERNAL_RFINTERACTIVE_EVALUATE = "robot.internal.rfinteractive.evaluate"  # Evaluate in Interactive Console
ROBOT_INTERNAL_RFINTERACTIVE_STOP = "robot.internal.rfinteractive.stop"  # Stop Interactive Console
ROBOT_INTERNAL_RFINTERACTIVE_SEMANTIC_TOKENS = "robot.internal.rfinteractive.semanticTokens"  # Get the semantic tokens based on the code entered.
ROBOT_INTERNAL_RFINTERACTIVE_RESOLVE_COMPLETION = "robot.internal.rfinteractive.resolveCompletion"  # Resolves the passed completion.
ROBOT_INTERNAL_RFINTERACTIVE_COMPLETIONS = "robot.internal.rfinteractive.completions"  # Get the completions based on the code entered.
ROBOT_GET_RFLS_HOME_DIR = "robot.getRFLSHomeDir"  # Provides the directory used to store information (usually ~/.robotframework-ls -- may be configured with `ROBOTFRAMEWORK_LS_USER_HOME` environment variable).
ROBOT_CLEAR_CACHES_AND_RESTART_PROCESSES = "robot.clearCachesAndRestartProcesses"  # Clear caches and restart Robot Framework Language Server
ROBOT_CLEAR_CACHES_AND_RESTART_PROCESSES_START_INTERNAL = "robot.clearCachesAndRestartProcesses.start.internal"  # Stops the RFLS and waits for robot.clearCachesAndRestartProcesses.finish.internal to restart
ROBOT_CLEAR_CACHES_AND_RESTART_PROCESSES_FINISH_INTERNAL = "robot.clearCachesAndRestartProcesses.finish.internal"  # To be used to restart the processes
ROBOT_START_INDEXING_INTERNAL = "robot.startIndexing.internal"  # Starts the indexing service
ROBOT_WAIT_FULL_TEST_COLLECTION_INTERNAL = "robot.waitFullTestCollection.internal"  # Schedules and Waits for a full test collection
ROBOT_RF_INFO_INTERNAL = "robot.rfInfo.internal"  # Collects information on the Robot Framework version being used.
ROBOT_LINT_WORKSPACE = "robot.lint.workspace"  # Lint all files in the workspace.
ROBOT_LINT_EXPLORER = "robot.lint.explorer"  # Lint
ROBOT_OPEN_FLOW_EXPLORER = "robot.openFlowExplorer"  # Open Robot Flow Explorer
ROBOT_OPEN_FLOW_EXPLORER_INTERNAL = "robot.openFlowExplorer.internal"  # Open Robot Flow Explorer Internals
ROBOT_GENERATE_FLOW_EXPLORER_MODEL = "robot.generateFlowExplorerModel"  # Generate Robot Flow Explorer Model
ROBOT_COLLECT_ROBOT_DOCUMENTATION = "robot.collectRobotDocumentation"  # Collect Robot Documentation
ROBOT_VIEW_DOCUMENTATION_PIN = "robot.view.documentation.pin"  # Pin Robot Documentation
ROBOT_VIEW_DOCUMENTATION_UNPIN = "robot.view.documentation.unpin"  # Unpin Robot Documentation
ROBOT_CONVERT_OUTPUT_XML_TO_ROBOSTREAM = "robot.convertOutputXMLToRobostream"  # Convert output.xml to robocorp stream
ROBOT_APPLY_CODE_ACTION = "robot.applyCodeAction"  # Applies some code action

ALL_SERVER_COMMANDS: List[str] = [
    ROBOT_INTERNAL_RFINTERACTIVE_START,
    ROBOT_INTERNAL_RFINTERACTIVE_EVALUATE,
    ROBOT_INTERNAL_RFINTERACTIVE_STOP,
    ROBOT_INTERNAL_RFINTERACTIVE_SEMANTIC_TOKENS,
    ROBOT_INTERNAL_RFINTERACTIVE_RESOLVE_COMPLETION,
    ROBOT_INTERNAL_RFINTERACTIVE_COMPLETIONS,
    ROBOT_GET_RFLS_HOME_DIR,
    ROBOT_START_INDEXING_INTERNAL,
    ROBOT_WAIT_FULL_TEST_COLLECTION_INTERNAL,
    ROBOT_RF_INFO_INTERNAL,
    ROBOT_LINT_WORKSPACE,
    ROBOT_LINT_EXPLORER,
    ROBOT_OPEN_FLOW_EXPLORER_INTERNAL,
    ROBOT_GENERATE_FLOW_EXPLORER_MODEL,
    ROBOT_COLLECT_ROBOT_DOCUMENTATION,
    ROBOT_CONVERT_OUTPUT_XML_TO_ROBOSTREAM,
    ROBOT_APPLY_CODE_ACTION,
]

# fmt: on



import functools
import sys

try:
    import rich_click as click
except ImportError:
    import click

from robotidy.exceptions import RobotidyConfigError


def catch_exceptions(func):
    """
    Catch exceptions and print user friendly message for common issues
    """

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        if not func:
            return functools.partial(catch_exceptions)
        try:
            return func(*args, **kwargs)
        except (RobotidyConfigError) as err:
            print(f"Error: {err}")
            sys.exit(1)
        except (click.exceptions.ClickException, click.exceptions.Exit):
            raise
        except Exception as err:
            message = (
                "\nFatal exception occurred. You can create an issue at "
                "https://github.com/MarketSquare/robotframework-tidy/issues . Thanks!"
            )
            err.args = (err.args[0] + message,) + err.args[1:]
            raise err

    return wrapper



from robotidy.utils import normalize_name


class RunKeywordVariant:
    def __init__(self, libname, name, resolve=1, branches=None, split_on_and=False):
        self.libname = normalize_name(libname)
        self.name = normalize_name(name)
        self.resolve = resolve
        self.branches = branches
        self.split_on_and = split_on_and


_RUN_KW = [
    RunKeywordVariant("BuiltIn", "Run Keyword"),
    RunKeywordVariant("BuiltIn", "Run Keyword And Continue On Failure"),
    RunKeywordVariant("BuiltIn", "Run Keyword And Expect Error", resolve=2),
    RunKeywordVariant("BuiltIn", "Run Keyword And Ignore Error"),
    RunKeywordVariant("BuiltIn", "Run Keyword And Return"),
    RunKeywordVariant("BuiltIn", "Run Keyword And Return If", resolve=2),
    RunKeywordVariant("BuiltIn", "Run Keyword And Return Status"),
    RunKeywordVariant("BuiltIn", "Run Keyword And Warn On Failure"),
    RunKeywordVariant("BuiltIn", "Run Keyword If", resolve=2, branches=["ELSE IF", "ELSE"]),
    RunKeywordVariant("BuiltIn", "Run Keyword If All Tests Passed"),
    RunKeywordVariant("BuiltIn", "Run Keyword If Any Tests Failed"),
    RunKeywordVariant("BuiltIn", "Run Keyword If Test Failed"),
    RunKeywordVariant("BuiltIn", "Run Keyword If Test Passed"),
    RunKeywordVariant("BuiltIn", "Run Keyword If Timeout Occurred"),
    RunKeywordVariant("BuiltIn", "Run Keyword Unless", resolve=2),
    RunKeywordVariant("BuiltIn", "Run Keywords", split_on_and=True),
    RunKeywordVariant("BuiltIn", "Repeat Keyword", resolve=2),
    RunKeywordVariant("BuiltIn", "Wait Until Keyword Succeeds", resolve=3),
]


def get_run_keywords():
    run_keywords = dict()
    for run_kw in _RUN_KW:
        run_keywords[run_kw.name] = run_kw
        run_keywords[f"{run_kw.libname}.{run_kw.name}"] = run_kw
    return run_keywords



import os
from collections import defaultdict
from pathlib import Path
from typing import Dict, List, Optional, Pattern, Tuple

from robotidy.files import get_paths
from robotidy.transformers import load_transformers


class FormattingConfig:
    def __init__(
        self,
        space_count: int,
        indent: Optional[int],
        continuation_indent: Optional[int],
        line_sep: str,
        start_line: Optional[int],
        end_line: Optional[int],
        separator: str,
        line_length: int,
    ):
        self.start_line = start_line
        self.end_line = end_line
        self.space_count = space_count
        self.line_length = line_length

        if indent is None:
            indent = space_count
        if continuation_indent is None:
            continuation_indent = space_count

        if separator == "space":
            self.separator = " " * space_count
            self.indent = " " * indent
            self.continuation_indent = " " * continuation_indent
        elif separator == "tab":
            self.separator = "\t"
            self.indent = "\t"
            self.continuation_indent = "\t"

        self.line_sep = self.get_line_sep(line_sep)

    @staticmethod
    def get_line_sep(line_sep):
        if line_sep == "windows":
            return "\r\n"
        elif line_sep == "unix":
            return "\n"
        elif line_sep == "auto":
            return "auto"
        else:
            return os.linesep


class Config:
    def __init__(
        self,
        formatting: FormattingConfig,
        skip,
        transformers: List[Tuple[str, List]],
        transformers_config: List[Tuple[str, List]],
        src: Tuple[str, ...],
        exclude: Optional[Pattern],
        extend_exclude: Optional[Pattern],
        skip_gitignore: bool,
        overwrite: bool,
        show_diff: bool,
        verbose: bool,
        check: bool,
        output: Optional[Path],
        force_order: bool,
        target_version: int,
        color: bool,
    ):
        self.sources = get_paths(src, exclude, extend_exclude, skip_gitignore)
        self.formatting = formatting
        self.overwrite = overwrite
        self.show_diff = show_diff
        self.verbose = verbose
        self.check = check
        self.output = output
        self.color = color
        transformers_config = self.convert_configure(transformers_config)
        self.transformers = self.get_transformers_instances(
            transformers, transformers_config, force_order, target_version, skip
        )
        transformer_map = {transformer.__class__.__name__: transformer for transformer in self.transformers}
        for transformer in self.transformers:
            # inject global settings TODO: handle it better
            setattr(transformer, "formatting_config", self.formatting)
            setattr(transformer, "transformers", transformer_map)

    @staticmethod
    def get_transformers_instances(transformers, transformers_config, force_order, target_version, skip):
        return [
            transformer.instance
            for transformer in load_transformers(
                transformers, transformers_config, force_order=force_order, target_version=target_version, skip=skip
            )
        ]

    @staticmethod
    def convert_configure(configure: List[Tuple[str, List]]) -> Dict[str, List]:
        config_map = defaultdict(list)
        for transformer, args in configure:
            config_map[transformer].extend(args)
        return config_map



def _get_robot_naked_version():
    import robot  # noqa

    return str(robot.get_version(True))


_found_major_version = None


def get_robot_major_version() -> int:
    global _found_major_version
    if _found_major_version is not None:
        return _found_major_version

    robot_version = _get_robot_naked_version()
    major_version = int(robot_version.split(".")[0])
    _found_major_version = major_version
    return major_version



from typing import Tuple, List, Optional, Callable, Any

from robocorp_ls_core.robotframework_log import get_logger
from robotframework_ls.impl.protocols import (
    ILocalizationInfo,
    ICompletionContext,
    INode,
)
from robotframework_ls.impl.robot_version import (
    get_robot_major_minor_version,
)


log = get_logger(__name__)


class Section(object):

    # Header to the section (i.e.: *** Setting ***)
    markers: Tuple[str, ...] = ()

    # Names that can appear under the section.
    @classmethod
    def get_names_in_section_pre_rf_5_1(cls) -> Tuple[str, ...]:
        raise NotImplementedError()

    # The names that can appear multiple times.
    multi_use: Tuple[str, ...] = ("Metadata", "Library", "Resource", "Variables")

    names_in_brackets: bool = True

    @classmethod
    def markers_for_lang(cls, lang):
        raise NotImplementedError()

    @classmethod
    def names_for_lang(cls, lang):
        raise NotImplementedError()


class _TestCaseFileSettingsSection(Section):
    matches_rf_node_class = "SettingSection"

    markers = ("Setting", "Settings")
    names_in_brackets = False

    @classmethod
    def get_names_in_section_pre_rf_5_1(cls) -> Tuple[str, ...]:
        return (
            "Documentation",
            "Metadata",
            "Suite Setup",
            "Suite Teardown",
            "Test Setup",
            "Test Teardown",
            "Test Template",
            "Test Timeout",
            "Library",
            "Resource",
            "Variables",
            "Force Tags",
            "Default Tags",
            # Task (aliases for Test).
            "Task Setup",
            "Task Teardown",
            "Task Template",
            "Task Timeout",
        )

    @classmethod
    def markers_for_lang(cls, lang):
        return (lang.settings_header,)

    @classmethod
    def names_for_lang(cls, lang):
        # Only called in RF 5.1 onwards (so, we don't need to show deprecated names).
        return (
            lang.documentation_setting,
            lang.metadata_setting,
            lang.suite_setup_setting,
            lang.suite_teardown_setting,
            lang.test_setup_setting,
            lang.task_setup_setting,
            lang.test_teardown_setting,
            lang.task_teardown_setting,
            lang.test_template_setting,
            lang.task_template_setting,
            lang.test_timeout_setting,
            lang.task_timeout_setting,
            lang.test_tags_setting,  # In place of Force Tags
            lang.task_tags_setting,  # In place of Force Tags
            # "Force Tags",  # Deprecated: https://github.com/robotframework/robotframework/issues/4365
            # "Default Tags",  # Deprecated: https://github.com/robotframework/robotframework/issues/4365
            lang.keyword_tags_setting,  # New: https://github.com/robotframework/robotframework/issues/4373
            lang.library_setting,
            lang.resource_setting,
            lang.variables_setting,
        )


class _InitFileSettingsSection(Section):
    matches_rf_node_class = "SettingSection"
    markers = ("Setting", "Settings")
    names_in_brackets = False

    @classmethod
    def get_names_in_section_pre_rf_5_1(cls) -> Tuple[str, ...]:
        names: Tuple[str, ...] = (
            "Documentation",
            "Metadata",
            "Suite Setup",
            "Suite Teardown",
            "Test Setup",
            "Test Teardown",
            "Test Timeout",
            "Library",
            "Resource",
            "Variables",
        )

        if get_robot_major_minor_version() >= (5, 1):
            names += ("Test Tags",)
        else:
            names += ("Force Tags",)
        return names

    @classmethod
    def markers_for_lang(cls, lang):
        return (lang.settings_header,)

    @classmethod
    def names_for_lang(cls, lang):
        # Only called in RF 5.1 onwards (so, we don't need to show deprecated names).
        return (
            lang.documentation_setting,
            lang.metadata_setting,
            lang.suite_setup_setting,
            lang.suite_teardown_setting,
            lang.test_setup_setting,
            lang.task_setup_setting,
            lang.test_teardown_setting,
            lang.task_teardown_setting,
            lang.test_timeout_setting,
            lang.task_timeout_setting,
            lang.test_tags_setting,  # In place of Force Tags
            lang.task_tags_setting,  # In place of Force Tags
            # "Force Tags",  # Deprecated: https://github.com/robotframework/robotframework/issues/4365
            lang.keyword_tags_setting,  # New: https://github.com/robotframework/robotframework/issues/4373
            lang.library_setting,
            lang.resource_setting,
            lang.variables_setting,
        )


class _ResourceFileSettingsSection(Section):
    matches_rf_node_class = "SettingSection"
    markers = ("Setting", "Settings")
    names_in_brackets = False

    @classmethod
    def get_names_in_section_pre_rf_5_1(cls) -> Tuple[str, ...]:
        return ("Documentation", "Library", "Resource", "Variables")

    @classmethod
    def markers_for_lang(cls, lang):
        return (lang.settings_header,)

    @classmethod
    def names_for_lang(cls, lang):
        return (
            lang.documentation_setting,
            lang.library_setting,
            lang.resource_setting,
            lang.variables_setting,
            lang.keyword_tags_setting,  # New: https://github.com/robotframework/robotframework/issues/4373
        )


class _VariableSection(Section):
    matches_rf_node_class = "VariableSection"
    markers = ("Variables", "Variable")

    @classmethod
    def get_names_in_section_pre_rf_5_1(cls) -> Tuple[str, ...]:
        return ()

    @classmethod
    def markers_for_lang(cls, lang):
        return (lang.variables_header,)


class _TestCaseSection(Section):
    matches_rf_node_class = "TestCaseSection"
    markers = ("Test Cases", "Test Case", "Tasks", "Task")

    @classmethod
    def get_names_in_section_pre_rf_5_1(cls) -> Tuple[str, ...]:
        return ("Documentation", "Tags", "Setup", "Teardown", "Template", "Timeout")

    @classmethod
    def markers_for_lang(cls, lang):
        return (lang.tasks_header, lang.test_cases_header)

    @classmethod
    def names_for_lang(cls, lang):
        return (
            lang.documentation_setting,
            lang.tags_setting,
            lang.setup_setting,
            lang.teardown_setting,
            lang.template_setting,
            lang.timeout_setting,
        )


class _KeywordSection(Section):
    matches_rf_node_class = "KeywordSection"
    markers = ("Keywords", "Keyword")

    @classmethod
    def get_names_in_section_pre_rf_5_1(cls) -> Tuple[str, ...]:
        return ("Documentation", "Arguments", "Teardown", "Timeout", "Tags", "Return")

    @classmethod
    def markers_for_lang(cls, lang):
        return (lang.keywords_header,)

    @classmethod
    def names_for_lang(cls, lang):
        return (
            lang.documentation_setting,
            lang.arguments_setting,
            lang.teardown_setting,
            lang.timeout_setting,
            lang.tags_setting,
            "Return",  # No translation
        )


class _CommentSection(Section):
    matches_rf_node_class = "CommentSection"
    markers = ("Comment", "Comments")

    @classmethod
    def get_names_in_section_pre_rf_5_1(cls) -> Tuple[str, ...]:
        return ()

    @classmethod
    def markers_for_lang(cls, lang):
        return (lang.comments_header,)

    @classmethod
    def names_for_lang(cls, lang):
        return ()


# Sections that can appear in a test case
_TEST_CASE_FILE_SECTIONS = [
    _TestCaseFileSettingsSection,
    _VariableSection,
    _TestCaseSection,
    _KeywordSection,
    _CommentSection,
]

# Sections that can appear in an init file
_INIT_FILE_SECTIONS = [
    _InitFileSettingsSection,
    _VariableSection,
    _KeywordSection,
    _CommentSection,
]

# Sections that can appear in a resource file
_RESOURCE_FILE_SECTIONS = [
    _ResourceFileSettingsSection,
    _VariableSection,
    _KeywordSection,
    _CommentSection,
]


def _get_accepted_section_header_words(
    completion_context: ICompletionContext, accept_word: Callable[[Any, str], bool]
) -> List[str]:
    from robotframework_ls.impl.robot_version import robot_version_supports_language

    sections = _get_accepted_sections(completion_context)
    ret = []
    if robot_version_supports_language():

        from robot.api import Language

        locinfo: ILocalizationInfo = completion_context.get_ast_localization_info()

        for section in sections:
            lang: Language
            for lang in locinfo.iter_languages_on_write():
                markers = section.markers_for_lang(lang)

                for marker in markers:
                    if accept_word(lang, marker):
                        ret.append(marker.title())
    else:
        for section in sections:
            for marker in section.markers:
                if accept_word(None, marker):
                    ret.append(marker.title())

    ret.sort()
    return ret


def _get_accepted_sections(completion_context) -> list:
    t = completion_context.get_type()
    if t == completion_context.TYPE_TEST_CASE:
        return _TEST_CASE_FILE_SECTIONS

    elif t == completion_context.TYPE_RESOURCE:
        return _RESOURCE_FILE_SECTIONS

    elif t == completion_context.TYPE_INIT:
        return _INIT_FILE_SECTIONS

    else:
        log.critical("Unrecognized section: %s", t)
        return _TEST_CASE_FILE_SECTIONS


def get_section_constant(completion_context, section: INode) -> Optional[Section]:
    accepted_sections = _get_accepted_sections(completion_context)

    for s in accepted_sections:
        if s.matches_rf_node_class == section.__class__.__name__:
            return s
    return None


# Default: accept only plural
def _accept_word_default(language, word):
    # Language is None on RF < 5.1
    if language is None:
        return word.endswith("s")

    # On RF 5.1 onwards the translation only requests translation
    # for the plural version. Note that this effectively removes
    # support for this setting on RF 5.1.
    return True


def complete(completion_context: ICompletionContext):
    from robocorp_ls_core.lsp import CompletionItemKind
    from robocorp_ls_core.lsp import CompletionItem
    from robocorp_ls_core.lsp import TextEdit
    from robocorp_ls_core.lsp import Range
    from robocorp_ls_core.lsp import Position
    from robotframework_ls.impl import text_utilities
    from robotframework_ls.impl.string_matcher import RobotStringMatcher
    from robotframework_ls.impl.robot_lsp_constants import (
        OPTION_ROBOT_COMPLETION_SECTION_HEADERS_FORM,
    )
    from robotframework_ls.impl.robot_lsp_constants import (
        OPTION_ROBOT_COMPLETION_SECTION_HEADERS_FORM_PLURAL,
    )

    selection = completion_context.sel  #: :type selection: DocumentSelection
    line_start = selection.line_to_column
    items = []

    if line_start:
        tu = text_utilities.TextUtilities(line_start)

        if tu.strip_leading_chars("*"):  # i.e.: the line must start with '*'
            line_to_end = selection.line_to_end
            tu.strip()

            # On RF 5.1 onwards the translation only requests translation
            # for the plural version. Note that this effectively removes
            # support for this setting on RF 5.1.
            # We should deprecate it in the future...
            form = OPTION_ROBOT_COMPLETION_SECTION_HEADERS_FORM_PLURAL
            config = completion_context.config
            if config is not None:
                form = config.get_setting(
                    OPTION_ROBOT_COMPLETION_SECTION_HEADERS_FORM,
                    str,
                    OPTION_ROBOT_COMPLETION_SECTION_HEADERS_FORM_PLURAL,
                )

            accept_word = _accept_word_default
            if form == "singular":

                def accept_word(language, word):  # noqa
                    if language is None:
                        return not word.endswith("s")
                    return True

            elif form == "both":

                def accept_word(language, word):  # noqa
                    return True

            words = _get_accepted_section_header_words(completion_context, accept_word)
            matcher = RobotStringMatcher(tu.text)

            # If we have a *, replace until that char (with it included)
            delta_to_replace = line_to_end.rfind("*") + 1
            if delta_to_replace == 0:
                for i, c in enumerate(line_to_end):
                    if c == "#":
                        # We want to preserve breaks
                        break

                    if c in (" ", "\t"):
                        continue

                    delta_to_replace = i + 1  # +1 to replace the char we just saw

            for word in words:
                if matcher.accepts(word):
                    label = "*** %s ***" % (word,)

                    text_edit = TextEdit(
                        Range(
                            # i.e.: always replace from the start of the line.
                            start=Position(selection.line, 0),
                            end=Position(
                                selection.line, selection.col + delta_to_replace
                            ),
                        ),
                        label,
                    )
                    # text_edit = None
                    items.append(
                        CompletionItem(
                            label,
                            kind=CompletionItemKind.Class,
                            text_edit=text_edit,
                            insertText=text_edit.newText,
                        )
                    )

    return [item.to_dict() for item in items]



"""
Methods for transforming Robot Framework ast model programmatically.
"""
from typing import Optional

from robotidy.app import Robotidy
from robotidy.cli import TransformType, find_and_read_config, validate_regex
from robotidy.config import Config, FormattingConfig
from robotidy.disablers import RegisterDisablers
from robotidy.files import DEFAULT_EXCLUDES
from robotidy.skip import SkipConfig
from robotidy.utils import ROBOT_VERSION


def get_skip_config(config):
    skip_documentation = config.get("skip_documentation", False)
    skip_return_values = config.get("skip_return_values", False)
    skip_keyword_call = config.get("skip_keyword_call", [])
    skip_keyword_call_pattern = config.get("skip_keyword_call_pattern", [])
    skip_settings = config.get("skip_settings", False)
    skip_arguments = config.get("skip_arguments", False)
    skip_setup = config.get("skip_setup", False)
    skip_teardown = config.get("skip_teardown", False)
    skip_template = config.get("skip_template", False)
    skip_timeout = config.get("skip_timeout", False)
    skip_return = config.get("skip_return", False)
    skip_tags = config.get("skip_tags", False)
    return SkipConfig(
        documentation=skip_documentation,
        return_values=skip_return_values,
        keyword_call=skip_keyword_call,
        keyword_call_pattern=skip_keyword_call_pattern,
        settings=skip_settings,
        arguments=skip_arguments,
        setup=skip_setup,
        teardown=skip_teardown,
        template=skip_template,
        timeout=skip_timeout,
        return_statement=skip_return,
        tags=skip_tags,
    )


def get_formatting_config(config, kwargs):
    space_count = kwargs.get("spacecount", None) or int(config.get("spacecount", 4))
    indent = kwargs.get("indent", None) or int(config.get("indent", space_count))
    cont_indent = kwargs.get("continuation_indent", None) or int(config.get("continuation_indent", space_count))
    formatting_config = FormattingConfig(
        space_count=space_count,
        indent=indent,
        continuation_indent=cont_indent,
        separator=kwargs.get("separator", None) or config.get("separator", "space"),
        line_sep=config.get("lineseparator", "native"),
        start_line=kwargs.get("startline", None) or int(config["startline"]) if "startline" in config else None,
        end_line=kwargs.get("endline", None) or int(config["endline"]) if "endline" in config else None,
        line_length=kwargs.get("line_length", None) or int(config.get("line_length", 120)),
    )
    return formatting_config


def get_robotidy(src: str, output: Optional[str], **kwargs):
    # TODO Refactor - Config should be read in one place both for API and CLI
    # TODO Remove kwargs usage - other SDKs are not using this feature
    config = find_and_read_config((src,))
    config = {k: str(v) if not isinstance(v, (list, dict)) else v for k, v in config.items()}
    converter = TransformType()
    transformers = [converter.convert(tr, None, None) for tr in config.get("transform", ())]
    configurations = [converter.convert(c, None, None) for c in config.get("configure", ())]
    formatting_config = get_formatting_config(config, kwargs)
    exclude = config.get("exclude", None)
    extend_exclude = config.get("extend_exclude", None)
    exclude = validate_regex(exclude if exclude is not None else DEFAULT_EXCLUDES)
    extend_exclude = validate_regex(extend_exclude)
    global_skip = get_skip_config(config)
    configuration = Config(
        transformers=transformers,
        transformers_config=configurations,
        skip=global_skip,
        src=(),
        exclude=exclude,
        extend_exclude=extend_exclude,
        skip_gitignore=False,
        overwrite=False,
        show_diff=False,
        formatting=formatting_config,
        verbose=False,
        check=False,
        output=output,
        force_order=False,
        target_version=ROBOT_VERSION.major,
        color=False,
    )
    return Robotidy(config=configuration)


def transform_model(model, root_dir: str, output: Optional[str] = None, **kwargs) -> Optional[str]:
    """
    :param model: The model to be transformed.
    :param root_dir: Root directory. Configuration file is searched based
    on this directory or one of its parents.
    :param output: Path where transformed model should be saved
    :param kwargs: Default values for global formatting parameters
    such as ``spacecount``, ``startline`` and ``endline``.
    :return: The transformed model converted to string or None if no transformation took place.
    """
    robotidy_class = get_robotidy(root_dir, output, **kwargs)
    disabler_finder = RegisterDisablers(
        robotidy_class.config.formatting.start_line, robotidy_class.config.formatting.end_line
    )
    disabler_finder.visit(model)
    if disabler_finder.file_disabled:
        return None
    diff, _, new_model = robotidy_class.transform(model, disabler_finder.disablers)
    if not diff:
        return None
    return new_model.text



from typing import List, Container

from robocorp_ls_core.protocols import check_implements, IDocumentSelection
from robocorp_ls_core.robotframework_log import get_logger
from robotframework_ls.impl.protocols import (
    ICompletionContext,
    IKeywordFound,
    IKeywordCollector,
    AbstractKeywordCollector,
    KeywordUsageInfo,
)
from robocorp_ls_core.lsp import (
    TextEditTypedDict,
    CompletionItemTypedDict,
    InsertTextFormat,
)
from robotframework_ls.impl.text_utilities import normalize_robot_name


log = get_logger(__name__)


class _Collector(AbstractKeywordCollector):
    def __init__(
        self, completion_context: ICompletionContext, keyword_usage: KeywordUsageInfo
    ):
        from robotframework_ls.impl.string_matcher import RobotStringMatcher
        from robotframework_ls.impl.string_matcher import (
            build_matchers_with_resource_or_library_scope,
        )
        from robotframework_ls.robot_config import create_convert_keyword_format_func
        from robot.api import Token
        from robotframework_ls.impl.robot_lsp_constants import (
            OPTION_ROBOT_COMPLETION_KEYWORDS_ARGUMENTS_SEPARATOR,
            OPTION_ROBOT_COMPLETIONS_KEYWORDS_PREFIX_IMPORT_NAME,
            OPTION_ROBOT_COMPLETIONS_KEYWORDS_PREFIX_IMPORT_NAME_IGNORE,
        )

        token_str = keyword_usage.token.value

        self.completion_items: List[CompletionItemTypedDict] = []
        self.completion_context = completion_context
        self.selection = completion_context.sel
        self.token = keyword_usage.token

        self._add_arguments = True
        if keyword_usage.node.type in (Token.TEMPLATE, Token.TEST_TEMPLATE):
            # i.e.: In templates the arguments are added in the test.
            self._add_arguments = False
        else:
            # i.e.: if we have arguments already don't add arguments again.
            for t in keyword_usage.node.tokens:
                if t.type == t.ARGUMENT and t.value:
                    self._add_arguments = False
                    break

        config = completion_context.config
        self._arguments_separator = "    "
        self._prefix_with_import_name = False
        self._prefix_with_import_name_ignore: Container[str] = ()
        if config:
            self._arguments_separator = config.get_setting(
                OPTION_ROBOT_COMPLETION_KEYWORDS_ARGUMENTS_SEPARATOR, str, "    "
            ).replace(r"\t", "\t")

            self._prefix_with_import_name = config.get_setting(
                OPTION_ROBOT_COMPLETIONS_KEYWORDS_PREFIX_IMPORT_NAME, bool, False
            )

            self._prefix_with_import_name_ignore = set(
                normalize_robot_name(x)
                for x in config.get_setting(
                    OPTION_ROBOT_COMPLETIONS_KEYWORDS_PREFIX_IMPORT_NAME_IGNORE,
                    (list, tuple),
                    (),
                )
            )

        self._is_dotted_keyword_name = "." in token_str
        self._matcher = RobotStringMatcher(token_str)
        self._scope_matchers = build_matchers_with_resource_or_library_scope(token_str)
        config = completion_context.config

        self._convert_keyword_format = create_convert_keyword_format_func(config)

    def accepts(self, keyword_name: str) -> bool:
        if self._matcher.accepts_keyword_name(keyword_name):
            return True
        for matcher in self._scope_matchers:
            if matcher.accepts_keyword_name(keyword_name):
                return True
        return False

    def _create_completion_item_from_keyword(
        self,
        keyword_found: IKeywordFound,
        selection: IDocumentSelection,
        token,
        prefix_with_import_name: bool,
        col_delta: int = 0,
    ) -> CompletionItemTypedDict:
        from robotframework_ls.impl.protocols import IKeywordArg
        from robotframework_ls.impl.variable_resolve import iter_robot_variable_matches

        label = keyword_found.keyword_name

        if keyword_found.library_name:
            # If we found the keyword in a library, convert its format depending on
            # the user configuration.
            label = self._convert_keyword_format(label)

        text = label
        replace_idx = 0
        if "{" in text:
            new_text = []
            var_match = None
            for var_match, _ in iter_robot_variable_matches(text):
                new_text.append(var_match.before)
                if var_match.name:
                    replace_idx += 1
                    new_text.append(f"${{{replace_idx}:\\${var_match.base}}}")

            if var_match is not None:
                new_text.append(var_match.after)
                text = "".join(new_text)

        if self._add_arguments:
            arg: IKeywordArg
            for _i, arg in enumerate(keyword_found.keyword_args):
                if (
                    arg.is_keyword_arg
                    or arg.is_star_arg
                    or arg.default_value is not None
                ):
                    continue

                arg_name = arg.arg_name
                arg_name = (
                    arg_name.replace("$", "\\$").replace("{", "").replace("}", "")
                )

                replace_idx += 1
                text = f"{text}{self._arguments_separator}${{{replace_idx}:{arg_name}}}"

        if keyword_found.library_name:
            use_libname = keyword_found.library_alias
            if not use_libname:
                use_libname = keyword_found.library_name
            if prefix_with_import_name:

                if self._is_dotted_keyword_name:
                    # Note: the label must also be updated because we're
                    # going to replace from the start, so, the start
                    # from the label must also match as it'll be used
                    # in the filtering.
                    text = f"{use_libname}.{text}"
                    label = f"{use_libname}.{label}"
                else:
                    if (
                        normalize_robot_name(use_libname)
                        in self._prefix_with_import_name_ignore
                    ):
                        # Don't change the text...
                        label = f"{label} ({use_libname})"
                    else:
                        text = f"{use_libname}.{text}"
                        label = f"{label} ({use_libname})"
            else:
                label = f"{label} ({use_libname})"

        elif keyword_found.resource_name:
            if prefix_with_import_name:
                if self._is_dotted_keyword_name:
                    text = f"{keyword_found.resource_name}.{text}"
                    label = f"{keyword_found.resource_name}.{label}"
                else:
                    if (
                        normalize_robot_name(keyword_found.resource_name)
                        in self._prefix_with_import_name_ignore
                    ):
                        # Don't change the text...
                        label = f"{label} ({keyword_found.resource_name})"
                    else:
                        text = f"{keyword_found.resource_name}.{text}"
                        label = f"{label} ({keyword_found.resource_name})"
            else:
                label = f"{label} ({keyword_found.resource_name})"

        text_edit: TextEditTypedDict = {
            "range": {
                "start": {
                    "line": selection.line,
                    "character": token.col_offset + col_delta,
                },
                "end": {"line": selection.line, "character": token.end_col_offset},
            },
            "newText": text,
        }

        return {
            "label": label,
            "kind": keyword_found.completion_item_kind,
            "textEdit": text_edit,
            "insertText": text_edit["newText"],
            "insertTextFormat": InsertTextFormat.Snippet,
        }

    def on_keyword(self, keyword_found: IKeywordFound):
        col_delta = 0

        if not self._matcher.accepts_keyword_name(keyword_found.keyword_name):
            for matcher in self._scope_matchers:
                if matcher.accepts_keyword(keyword_found):
                    # +1 for the dot
                    col_delta = len(matcher.resource_or_library_name) + 1
                    break
            else:
                return  # i.e.: don't add completion

        prefix_with_import_name = self._prefix_with_import_name

        if prefix_with_import_name:
            if keyword_found.keyword_ast is not None:
                keyword_completion_context = keyword_found.completion_context
                if keyword_completion_context is not None:
                    if (
                        keyword_completion_context.doc.uri
                        == self.completion_context.doc.uri
                    ):
                        prefix_with_import_name = False

        if prefix_with_import_name:
            # If we're going to prefix with the module, replace
            # it completely and not from the dot.
            col_delta = 0

        item = self._create_completion_item_from_keyword(
            keyword_found,
            self.selection,
            self.token,
            prefix_with_import_name,
            col_delta=col_delta,
        )

        self.completion_context.assign_documentation_resolve(
            item, keyword_found.compute_docs_with_signature
        )

        self.completion_items.append(item)

    def __typecheckself__(self) -> None:
        _: IKeywordCollector = check_implements(self)


def complete(completion_context: ICompletionContext) -> List[CompletionItemTypedDict]:
    from robotframework_ls.impl.collect_keywords import collect_keywords
    from robotframework_ls.impl import ast_utils

    token_info = completion_context.get_current_token()
    if token_info is not None:
        keyword_usage = ast_utils.create_keyword_usage_info_from_token(
            token_info.stack, token_info.node, token_info.token
        )
        if keyword_usage is not None and not keyword_usage.token.value.strip().endswith(
            "}"
        ):
            collector = _Collector(completion_context, keyword_usage)
            collect_keywords(completion_context, collector)

            return collector.completion_items

    return []



"""
Reports are configurable summaries after Robocop scan. For example, it can be a total number of issues discovered.
They are dynamically loaded during setup according to a configuration.

Each report class collects rules messages from linter and parses it. At the end of the scan it will print the report.

To enable report use ``-r`` / ``--reports`` argument and the name of the report.
You can use separate arguments (``-r report1 -r report2``) or comma-separated list (``-r report1,report2``). Example::

    robocop --reports rules_by_id,some_other_report path/to/file.robot

To enable all default reports use ``--reports all``.

The order of the reports is preserved. For example, if you want ``timestamp`` report to be printed before any
other reports, you can use following configuration::

    robocop --reports timestamp,all src.robot

"""
import inspect
import json
import sys
from collections import OrderedDict, defaultdict
from datetime import datetime
from operator import itemgetter
from pathlib import Path
from timeit import default_timer as timer
from warnings import warn

import pytz
from dateutil import tz

import robocop.exceptions
from robocop.rules import Message
from robocop.utils import RecommendationFinder
from robocop.version import __version__


class Report:
    """
    Base class for report class.
    Override `configure` method if you want to allow report configuration.
    Override `add_message`` if your report processes the Robocop issues.

    Set class attribute `DEFAULT` to `False` if you don't want your report to be included in `all` reports.
    """

    DEFAULT = True

    def configure(self, name, value):
        raise robocop.exceptions.ConfigGeneralError(
            f"Provided param '{name}' for report '{getattr(self, 'name')}' does not exist"
        )  # noqa

    def add_message(self, *args):
        pass


def load_reports():
    """
    Load all valid reports.
    Report is considered valid if it inherits from `Report` class
    and contains both `name` and `description` attributes.
    """
    reports = {}
    classes = inspect.getmembers(sys.modules[__name__], inspect.isclass)
    for report_class in classes:
        if not issubclass(report_class[1], Report):
            continue
        report = report_class[1]()
        if not hasattr(report, "name") or not hasattr(report, "description"):
            continue
        reports[report.name] = report
    return reports


def is_report_default(report):
    return getattr(report, "DEFAULT", False)


def get_reports(configured_reports):
    """
    Returns dictionary with list of valid, enabled reports (listed in `configured_reports` set of str).
    If `configured_reports` contains `all` then all default reports are enabled.
    """
    reports = load_reports()
    enabled_reports = OrderedDict()
    for report in configured_reports:
        if report == "all":
            for name, report_class in reports.items():
                if is_report_default(report_class) and name not in enabled_reports:
                    enabled_reports[name] = report_class
        elif report not in reports:
            raise robocop.exceptions.InvalidReportName(report, reports)
        elif report not in enabled_reports:
            enabled_reports[report] = reports[report]
    return enabled_reports


def list_reports(reports):
    """Returns description of enabled reports."""
    sorted_by_name = sorted(reports.values(), key=lambda x: x.name)
    available_reports = "Available reports:\n"
    available_reports += "\n".join(f"{report.name:20} - {report.description}" for report in sorted_by_name) + "\n"
    available_reports += "all" + " " * 18 + "- Turns on all default reports"
    return available_reports


class RulesByIdReport(Report):
    """
    Report name: ``rules_by_id``

    Report that groups linter rules messages by rule id and prints it ordered by most common message.
    Example::

        Issues by ID:
        W0502 (too-little-calls-in-keyword) : 5
        W0201 (missing-doc-keyword)         : 4
        E0401 (parsing-error)               : 3
        W0301 (not-allowed-char-in-name)    : 2
        E0901 (keyword-after-return)        : 1
    """

    def __init__(self):
        self.name = "rules_by_id"
        self.description = "Groups detected issues by rule id and prints it ordered by most common"
        self.message_counter = defaultdict(int)

    def add_message(self, message: Message):  # noqa
        self.message_counter[message.get_fullname()] += 1

    def get_report(self) -> str:
        message_counter_ordered = sorted(self.message_counter.items(), key=itemgetter(1), reverse=True)
        report = "\nIssues by ID:\n"
        if not message_counter_ordered:
            report += "No issues found."
            return report
        longest_name = max(len(msg[0]) for msg in message_counter_ordered)
        report += "\n".join(f"{message:{longest_name}} : {count}" for message, count in message_counter_ordered)
        return report


class RulesBySeverityReport(Report):
    """
    Report name: ``rules_by_error_type``

    Report that groups linter rules messages by severity and prints total of issues per every severity level.

    Example::

        Found 15 issues: 11 WARNINGs, 4 ERRORs.
    """

    def __init__(self):
        self.name = "rules_by_error_type"
        self.description = "Prints total number of issues grouped by severity"
        self.severity_counter = defaultdict(int)

    def add_message(self, message: Message):
        self.severity_counter[message.severity] += 1

    def get_report(self) -> str:
        issues_count = sum(self.severity_counter.values())
        if not issues_count:
            return "\nFound 0 issues."

        report = "\nFound 1 issue: " if issues_count == 1 else f"\nFound {issues_count} issues: "
        warning_types = []
        for severity, count in self.severity_counter.items():
            plural = "" if count == 1 else "s"
            warning_types.append(f"{count} {severity.name}{plural}")
        report += ", ".join(warning_types)
        report += "."
        return report


class ReturnStatusReport(Report):
    """
    Report name: ``return_status``

    This report is always enabled.
    Report that checks if number of returned rules messages for given severity value does not exceed preset threshold.
    That information is later used as a return status from Robocop.
    """

    def __init__(self):
        self.name = "return_status"
        self.description = "Checks if number of specific issues exceed quality gate limits"
        self.return_status = 0
        self.counter = RulesBySeverityReport()
        self.quality_gate = {"E": 0, "W": 0, "I": -1}

    def configure(self, name, value):
        if name not in ["quality_gate", "quality_gates"]:
            super().configure(name, value)
        for val in value.split(":"):
            try:
                name, count = val.split("=", maxsplit=1)
                if name.upper() in self.quality_gate:
                    self.quality_gate[name.upper()] = int(count)
            except ValueError:
                continue

    def add_message(self, message: Message):
        self.counter.add_message(message)

    def get_report(self):
        for severity, count in self.counter.severity_counter.items():
            threshold = self.quality_gate.get(severity.value, 0)
            if -1 < threshold < count:
                self.return_status += count - threshold
        self.return_status = min(self.return_status, 255)


class TimeTakenReport(Report):
    """
    Report name: ``scan_timer``

    Report that returns Robocop execution time

    Example::

        Scan finished in 0.054s.
    """

    def __init__(self):
        self.name = "scan_timer"
        self.description = "Returns Robocop execution time"
        self.start_time = timer()

    def get_report(self) -> str:
        return f"\nScan finished in {timer() - self.start_time:.3f}s."


class JsonReport(Report):
    """
    Report name: ``json_report``

    Report that returns list of found issues in JSON format.
    """

    DEFAULT = False

    def __init__(self):
        self.name = "json_report"
        self.description = "Accumulates found issues in JSON format"
        self.issues = []

    def add_message(self, message: Message):
        self.issues.append(message.to_json())

    def get_report(self):
        return None


class FileStatsReport(Report):
    """
    Report name: ``file_stats``

    Report that displays overall statistics about number of processed files.

    Example::

        Processed 7 files from which 5 files contained issues.
    """

    def __init__(self):
        self.name = "file_stats"
        self.description = "Prints overall statistics about number of processed files"
        self.files_count = 0
        self.files_with_issues = set()

    def add_message(self, message: Message):
        self.files_with_issues.add(message.source)

    def get_report(self) -> str:
        if not self.files_count:
            return "\nNo files were processed."
        plural_files = "s" if self.files_count > 1 else ""
        if not self.files_with_issues:
            return f"\nProcessed {self.files_count} file{plural_files} but no issues were found."

        plural_files_with_issues = "" if len(self.files_with_issues) == 1 else "s"
        return (
            f"\nProcessed {self.files_count} file{plural_files} from which {len(self.files_with_issues)} "
            f"file{plural_files_with_issues} contained issues."
        )


class RobocopVersionReport(Report):
    """
    Report name: ``version``

    Report that returns Robocop version.

    Example::

        Report generated by Robocop version: 2.0.2
    """

    def __init__(self):
        self.name = "version"
        self.description = "Returns Robocop version"

    def get_report(self) -> str:
        return f"\nReport generated by Robocop version: {__version__}"


class TimestampReport(Report):
    """
    Report name: ``timestamp``

    Report that returns Robocop execution timestamp.
    Timestamp follows local time in format of
    `Year-Month-Day Hours(24-hour clock):Minutes:Seconds ±hh:mm UTC offset` as default.

    Example::

        Reported: 2022-07-10 21:25:00 +0300

    Both of default values, ``timezone`` and ``format`` can be configured by
    ``-c/--configure`` and ``timestamp:timezone:"<timezone name>"`` and/or ``timestamp:format:"<format string>"``::

        robocop --configure timestamp:timezone:"Europe/Paris" --configure timestamp:format:"%Y-%m-%d %H:%M:%S %Z %z"

    This yields following timestamp report::

         Reported: 2022-07-10 20:38:10 CEST +0200

    For timezone names,
    see: https://en.wikipedia.org/wiki/List_of_tz_database_time_zones

    For timestamp formats,
    see: https://docs.python.org/3/library/datetime.html#strftime-and-strptime-format-codes

    Useful configurations::

        Local time to ISO 8601 format:
        robocop --configure timestamp:format:"%Y-%m-%dT%H:%M:%S%z"

        UTC time:
        robocop --configure timestamp:timezone:"UTC" --configure timestamp:format:"%Y-%m-%dT%H:%M:%S %Z %z"

        Timestamp with high precision:
        robocop --configure timestamp:format:"%Y-%m-%dT%H:%M:%S.%f %z"

        12-hour clock:
        robocop --configure timestamp:format:"%Y-%m-%d %I:%M:%S %p %Z %z"

        More human readable format 'On 10 July 2022 07:26:24 +0300':
        robocop --configure timestamp:format:"On %d %B %Y %H:%M:%S %z"

    """

    def __init__(self):
        self.name = "timestamp"
        self.description = "Returns Robocop execution timestamp."
        self.timezone = "local"
        self.format = "%Y-%m-%d %H:%M:%S %z"

    def configure(self, name, value):
        if name == "timezone":
            self.timezone = value
        elif name == "format":
            if value:
                self.format = value
            else:
                warn("Empty format string for `timestamp` report does not make sense. Default format used.")
        else:
            super().configure(name, value)

    def get_report(self) -> str:
        return f"\nReported: {self._get_timestamp()}"

    def _get_timestamp(self) -> str:
        try:
            timezone = tz.tzlocal() if self.timezone == "local" else pytz.timezone(self.timezone)
            return datetime.now(timezone).strftime(self.format)
        except pytz.exceptions.UnknownTimeZoneError as err:
            raise robocop.exceptions.ConfigGeneralError(
                f"Provided timezone '{self.timezone}' for report '{getattr(self, 'name')}' is not valid. "
                "Use timezone names like `Europe\Helsinki`."
                "See: https://en.wikipedia.org/wiki/List_of_tz_database_time_zone"
            ) from err  # noqa


class SarifReport(Report):
    """
    Report name: ``sarif``

    Report that generates SARIF output file.

    This report is not included in the default reports. The ``--reports all`` option will not enable this report.
    You can still enable it using report name directly: ``--reports sarif`` or ``reports all,sarif``.

    All fields required by Github Code Scanning are supported. The output file will be generated
    in the current working directory with the ``.sarif.json`` name.

    You can configure output directory and report filename::

        robocop --configure sarif:output_dir=C:/sarif_reports --configure sarif:report_filename=.sarif

    """

    DEFAULT = False
    SCHEMA_VERSION = "2.1.0"
    SCHEMA = f"https://json.schemastore.org/sarif-{SCHEMA_VERSION}.json"

    def __init__(self):
        self.name = "sarif"
        self.description = "Generate SARIF output file"
        self.output_dir = None
        self.report_filename = ".sarif.json"
        self.issues = []

    def configure(self, name, value):
        if name == "output_dir":
            self.output_dir = Path(value)
            self.output_dir.mkdir(parents=True, exist_ok=True)
        elif name == "report_filename":
            self.report_filename = value
        else:
            super().configure(name, value)

    @staticmethod
    def map_severity_to_level(severity):
        return {"WARNING": "warning", "ERROR": "error", "INFO": "note"}[severity.name]

    def get_rule_desc(self, rule):
        return {
            "id": rule.rule_id,
            "name": rule.name,
            "helpUri": f"https://robocop.readthedocs.io/en/stable/rules.html#{rule.name}",
            "shortDescription": {"text": rule.msg},
            "fullDescription": {"text": rule.docs},
            "defaultConfiguration": {"level": self.map_severity_to_level(rule.default_severity)},
            "help": {"text": rule.docs, "markdown": rule.docs},
        }

    def add_message(self, message: Message):
        self.issues.append(message)

    def generate_sarif_issues(self, config):
        sarif_issues = []
        for issue in self.issues:
            relative_uri = Path(issue.source).relative_to(config.root)
            sarif_issue = {
                "ruleId": issue.rule_id,
                "level": self.map_severity_to_level(issue.severity),
                "message": {"text": issue.desc},
                "locations": [
                    {
                        "physicalLocation": {
                            "artifactLocation": {"uri": relative_uri.as_posix(), "uriBaseId": "%SRCROOT%"},
                            "region": {
                                "startLine": issue.line,
                                "endLine": issue.end_line,
                                "startColumn": issue.col,
                                "endColumn": issue.end_col,
                            },
                        }
                    }
                ],
            }
            sarif_issues.append(sarif_issue)
        return sarif_issues

    def generate_rules_config(self, rules):
        unique_enabled_rules = {rule.rule_id: rule for rule in rules.values() if rule.enabled}
        sorted_rules = sorted(unique_enabled_rules.values(), key=lambda x: x.rule_id)
        rules_config = [self.get_rule_desc(rule) for rule in sorted_rules]
        return rules_config

    def generate_sarif_report(self, config, rules):
        report = {
            "$schema": self.SCHEMA,
            "version": self.SCHEMA_VERSION,
            "runs": [
                {
                    "tool": {
                        "driver": {
                            "name": "Robocop",
                            "semanticVersion": __version__,
                            "informationUri": "https://robocop.readthedocs.io/",
                            "rules": self.generate_rules_config(rules),
                        }
                    },
                    "automationDetails": {"id": "robocop/"},
                    "results": self.generate_sarif_issues(config),
                }
            ],
        }
        return report

    def get_report(self, config, rules) -> str:
        report = self.generate_sarif_report(config, rules)
        if self.output_dir is not None:
            output_path = self.output_dir / self.report_filename
        else:
            output_path = Path(self.report_filename)
        with open(output_path, "w") as fp:
            json_string = json.dumps(report, indent=4)
            fp.write(json_string)
        return f"Generated sarif report in {output_path}"



from robot.api.parsing import Comment, CommentSection, EmptyLine

from robotidy.disablers import skip_section_if_disabled
from robotidy.transformers import Transformer


class DiscardEmptySections(Transformer):
    """
    Remove empty sections.
    Sections are considered empty if there are only empty lines inside.
    You can remove sections with only comments by setting ``allow_only_comments`` parameter to False:

    ```robotframework
    *** Variables ***
    # this section will be removed with ``alow_only_comments`` parameter set to False
    ```
    """

    def __init__(self, allow_only_comments: bool = True):
        super().__init__()
        # If False then sections with only with comments are considered to be empty
        self.allow_only_comments = allow_only_comments

    @skip_section_if_disabled
    def visit_Section(self, node):  # noqa
        anything_but = (
            EmptyLine if self.allow_only_comments or isinstance(node, CommentSection) else (Comment, EmptyLine)
        )
        if all(isinstance(child, anything_but) for child in node.body):
            return None
        return node



from typing import List

from robocorp_ls_core.lsp import CompletionItemTypedDict
from robocorp_ls_core.lsp import TextEditTypedDict
from robocorp_ls_core.lsp import InsertTextFormat
from robocorp_ls_core.lsp import CompletionItemKind
from robotframework_ls.impl.protocols import ICompletionContext
from robotframework_ls.impl.protocols import ILibraryImportNode
import os


def _iter_import_names(completion_context):

    imported_libraries = completion_context.get_imported_libraries()
    lib: ILibraryImportNode
    for lib in imported_libraries:
        alias = lib.alias
        if alias:
            use = alias
        else:
            use = lib.name
        if use.endswith(".py"):
            use = use[:-3]

        if use:
            yield use

    for resource_import in completion_context.get_resource_imports():
        use = resource_import.name
        use = os.path.splitext(use)[0]
        if use:
            yield use


def complete(completion_context: ICompletionContext) -> List[CompletionItemTypedDict]:
    from robotframework_ls.impl import ast_utils
    from robotframework_ls.impl.text_utilities import normalize_robot_name

    ret: List[CompletionItemTypedDict] = []
    token_info = completion_context.get_current_token()
    if token_info is not None:
        keyword_usage = ast_utils.create_keyword_usage_info_from_token(
            token_info.stack, token_info.node, token_info.token
        )
        if keyword_usage is not None and "{" not in keyword_usage.token.value:
            full_tok_name = keyword_usage.token.value
            diff = token_info.token.end_col_offset - completion_context.sel.col
            if diff < 0:
                return ret
            if diff > 0:
                full_tok_name = full_tok_name[:-diff]

            replace_up_to_col = completion_context.sel.col

            curr_normalized = normalize_robot_name(full_tok_name).replace(".", "")

            for use in _iter_import_names(completion_context):
                i = use.rfind("}")
                if i >= 0:
                    use = use[i + 1 :]

                use = os.path.basename(use)

                if curr_normalized in normalize_robot_name(use).replace(".", ""):
                    text_edit: TextEditTypedDict = {
                        "range": {
                            "start": {
                                "line": completion_context.sel.line,
                                "character": token_info.token.col_offset,
                            },
                            "end": {
                                "line": completion_context.sel.line,
                                "character": replace_up_to_col,
                            },
                        },
                        "newText": use,
                    }

                    label = use
                    ret.append(
                        {
                            "label": label,
                            "kind": CompletionItemKind.Module,
                            "textEdit": text_edit,
                            "insertText": text_edit["newText"],
                            "insertTextFormat": InsertTextFormat.Snippet,
                        }
                    )

    return ret



import xml.sax.handler
import sys
import datetime
from typing import Optional, Callable


class _RobotData:
    def __init__(self, attrs):
        self.attrs = attrs
        self.status = None

    def __str__(self):
        return f"_RobotData"

    __repr__ = __str__


class _SuiteData:
    def __init__(self, attrs):
        self.attrs = attrs
        self.status = None

    def __str__(self):
        return f'_SuiteData({self.attrs.get("name")})'

    __repr__ = __str__


class _TestData:
    def __init__(self, attrs):
        self.attrs = attrs
        self.status = None
        self.tags = []
        self.sent = False

    def __str__(self):
        return f'_TestData({self.attrs.get("name")})'

    __repr__ = __str__


class _AbstractKeywordData:
    def __init__(self, attrs):
        self.attrs = attrs
        self.status = None

        self.args = []

        self.doc = ""
        self.sent = False

    @property
    def name(self):
        raise NotImplementedError

    @property
    def type(self):
        raise NotImplementedError


class _KeywordData(_AbstractKeywordData):
    def __init__(self, attrs):
        _AbstractKeywordData.__init__(self, attrs)
        self.var_assign = []

    @property
    def name(self):
        return self.attrs.get("name")

    @property
    def type(self):
        return "KEYWORD"

    def __str__(self):
        return f"_KeywordData({self.name})"

    __repr__ = __str__


class _ForData(_AbstractKeywordData):
    def __init__(self, attrs):
        _AbstractKeywordData.__init__(self, attrs)
        self.var = "<unset>"
        self.values = []

    @property
    def flavor(self):
        return self.attrs.get("flavor", "<unknown-flavor>")

    @property
    def type(self):
        return "FOR"

    @property
    def name(self):
        return f"{self.var} {self.flavor} [{' | '.join(self.values)}]"

    def __str__(self):
        return f"_ForData({self.name})"

    __repr__ = __str__


class _ReturnData(_AbstractKeywordData):
    def __init__(self, attrs):
        _AbstractKeywordData.__init__(self, attrs)
        self.values = []

    @property
    def type(self):
        return "RETURN"

    @property
    def name(self):
        return f"{' | '.join(self.values)}"

    def __str__(self):
        return f"_ReturnData({self.name})"

    __repr__ = __str__


class _WhileData(_AbstractKeywordData):
    def __init__(self, attrs):
        _AbstractKeywordData.__init__(self, attrs)
        self.condition = attrs.get("condition", "<unknown condition>")

    @property
    def type(self):
        return "WHILE"

    @property
    def name(self):
        return f"{self.condition}"

    def __str__(self):
        return f"_WhileData({self.name})"

    __repr__ = __str__


class _BranchData(_AbstractKeywordData):
    """
    Branch data can be:
    - if/else if/else
    - try/except/finally
    """

    def __init__(self, attrs):
        _AbstractKeywordData.__init__(self, attrs)
        self.condition = attrs.get("condition")  # Used on if/else if
        self.pattern = None  # Used on except (message)

    @property
    def type(self):
        return self.attrs.get("type")

    @property
    def name(self):
        if self.condition:
            return self.condition
        if self.pattern:
            return self.pattern
        return ""

    def __str__(self):
        return f"_BranchData({self.name})"

    __repr__ = __str__


class _IterData(_AbstractKeywordData):
    def __init__(self, attrs):
        _AbstractKeywordData.__init__(self, attrs)
        self.var_name = None
        self.var = None

    @property
    def type(self):
        return "ITERATION"

    @property
    def name(self):
        if self.var_name is not None and self.var is not None:
            return f"{self.var_name} = {self.var}"
        if self.var_name is not None:
            return f"{self.var_name}"
        if self.var is not None:
            return f"{self.var}"
        return ""

    def __str__(self):
        return f"_IterData({self.name})"

    __repr__ = __str__


def compute_timedelta(initial_time: datetime.datetime, endtime_str: str):
    endtime: datetime.datetime = parse_time(endtime_str)
    delta = endtime - initial_time
    return round(delta.total_seconds(), 3)


class _Status:
    def __init__(self, attrs):
        self.attrs = attrs

    @property
    def status(self):
        return self.attrs["status"]

    @property
    def endtime(self):
        return self.attrs["endtime"]

    @property
    def starttime(self):
        return self.attrs["starttime"]

    def compute_start_timedelta(self, initial_time):
        return compute_timedelta(initial_time, self.starttime)

    def compute_end_timedelta(self, initial_time):
        return compute_timedelta(initial_time, self.endtime)


class _XmlSaxParser(xml.sax.handler.ContentHandler):
    """
    Notes:

    - Log messages are unscoped (they just appear as errors in the
      end and thus we can't know what was the current keyword).

    - The start/end time is written as the last thing in the
      element along with the status, so, we have a special message
      to write the start (in general the protocol would send it along
      with the test time, but we send it afterwards so that we can
      use the sax parser).
    """

    def __init__(self, create_listener):
        self._create_listener = create_listener
        self._listener = None
        self._stack = []
        self._need_chars = False
        self._chars = []
        self._curr_message_attrs = None
        self._found_logs = set()

    def startElement(self, name, attrs):
        method = getattr(self, "start_" + name, None)
        if method:
            try:
                method(attrs)
            except:
                raise RuntimeError(
                    f"Error when starting element: {name}. Attrs: {attrs}. Stack: {self._stack}"
                )
        # else:
        #     print("Unhandled start:", name)

    def endElement(self, name):
        method = getattr(self, "end_" + name, None)
        if method:
            try:
                method()
            except:
                raise RuntimeError(
                    f"Error when finishing element: {name}. Stack: {self._stack}"
                )
        # else:
        #     print("Unhandled end:", name)

    def start_robot(self, attrs):
        assert self._listener is None
        generator = attrs.get("generator")
        self._listener = self._create_listener(attrs)
        self._listener.send_info(generator)
        self._stack.append(_RobotData(attrs))

    def end_robot(self):
        self.send_delayed()
        del self._stack[-1]

    def start_suite(self, attrs):
        self.send_delayed()
        name = attrs.get("name")
        suiteid = attrs.get("id")
        source = attrs.get("source")
        if name and suiteid:
            self._stack.append(_SuiteData(attrs))
            self._listener.start_suite(
                name, {"id": suiteid, "source": source, "timedelta": -1}
            )
        else:
            self._stack.append(None)

    def end_suite(self):
        self.send_delayed()
        s = self._stack.pop(-1)
        if s is not None:
            status = s.status.attrs["status"]
            self._listener.send_start_time_delta(
                s.status.compute_start_timedelta(self._listener.initial_time)
            )

            self._listener.end_suite(
                s.attrs["name"],
                {"status": status, "timedelta": -1, "id": s.attrs["id"]},
            )

    def start_test(self, attrs):
        self.send_delayed()
        name = attrs.get("name")
        suiteid = attrs.get("id")
        if name and suiteid:
            self._stack.append(_TestData(attrs))
        else:
            self._stack.append(None)

    def end_test(self):
        self.send_delayed()
        s = self._stack.pop(-1)
        if s is not None:
            status = s.status.status

            self._listener.send_start_time_delta(
                s.status.compute_start_timedelta(self._listener.initial_time)
            )

            self._listener.end_test(
                s.attrs["name"],
                {
                    "id": s.attrs["id"],
                    "status": status,
                    "timedelta": s.status.compute_end_timedelta(
                        self._listener.initial_time
                    ),
                    "message": "",
                },
            )

    def start_kw(self, attrs):
        self.send_delayed()
        name = attrs.get("name")
        if name:
            self._stack.append(_KeywordData(attrs))

            # We can't send it right away because we need the args which will
            # just appear afterwards...
        else:
            self._stack.append(None)

    def send_delayed(self):
        if not self._stack:
            return

        peek = self._stack[-1]
        if not hasattr(peek, "sent"):
            return

        if not peek.sent:
            if isinstance(peek, _AbstractKeywordData):
                peek.sent = True
                attrs = peek.attrs
                name = peek.name
                libname = attrs.get("library", "")
                doc = peek.doc
                args = peek.args
                assign = getattr(peek, "var_assign", [])
                self._listener.start_keyword(
                    name,
                    {
                        "kwname": name,
                        "libname": libname,
                        "doc": doc,
                        "args": args,
                        "type": peek.type,
                        "timedelta": -1,
                        "assign": assign,
                        "source": None,  # source is never available in this case.
                        "lineno": -1,
                    },
                )
            elif isinstance(peek, _TestData):
                peek.sent = True

                attrs = peek.attrs
                name = attrs.get("name")
                suiteid = attrs.get("id")
                line = attrs.get("line")
                tags = peek.tags

                self._listener.start_test(
                    name, {"id": suiteid, "lineno": line, "timedelta": -1, "tags": tags}
                )

    def end_kw(self):
        self.send_delayed()
        s = self._stack.pop(-1)
        if s is not None:
            status = s.status.status

            self._listener.send_start_time_delta(
                s.status.compute_start_timedelta(self._listener.initial_time)
            )
            self._listener.end_keyword(
                s.name,
                {
                    "kwname": s.name,
                    "status": status,
                    "timedelta": s.status.compute_end_timedelta(
                        self._listener.initial_time
                    ),
                    "message": "",
                },
            )

    def _get_chars_and_disable(self):
        self._need_chars = False
        content = "".join(self._chars)
        self._chars = []
        return content

    def start_for(self, attrs):
        self.send_delayed()
        self._stack.append(_ForData(attrs))

    end_for = end_kw

    def start_return(self, attrs):
        self.send_delayed()
        self._stack.append(_ReturnData(attrs))

    end_return = end_kw

    def start_branch(self, attrs):
        self.send_delayed()
        self._stack.append(_BranchData(attrs))

    end_branch = end_kw

    def start_while(self, attrs):
        self.send_delayed()
        self._stack.append(_WhileData(attrs))

    end_while = end_kw

    def start_iter(self, attrs):
        self.send_delayed()
        self._stack.append(_IterData(attrs))

    end_iter = end_kw

    def start_var(self, attrs):
        self._need_chars = True
        data = self._stack[-1]
        if hasattr(data, "var_name"):
            data.var_name = attrs.get("name")

    def end_var(self):
        content = self._get_chars_and_disable()
        data = self._stack[-1]
        if hasattr(data, "var_assign"):
            data.var_assign.append(content)
        else:
            data.var = content

    def start_value(self, attrs):
        self._need_chars = True

    def end_value(self):
        content = self._get_chars_and_disable()
        data = self._stack[-1]
        data.values.append(content)

    def start_pattern(self, attrs):
        self._need_chars = True

    def end_pattern(self):
        content = self._get_chars_and_disable()
        data = self._stack[-1]
        if isinstance(data, _BranchData):
            data.pattern = content

    def start_arg(self, attrs):
        self._need_chars = True

    def end_arg(self):
        content = self._get_chars_and_disable()
        if self._stack:
            peek = self._stack[-1]
            if isinstance(peek, _AbstractKeywordData):
                peek.args.append(content)

    def start_tag(self, attrs):
        self._need_chars = True

    def end_tag(self):
        content = self._get_chars_and_disable()
        if self._stack:
            peek = self._stack[-1]
            if isinstance(peek, _TestData):
                if not peek.sent:
                    peek.tags.append(content)
                else:
                    # The test data was already sent (it can be sent
                    # just at the end, which isn't ideal...).
                    self.send_delayed()
                    self._listener.send_tag(content)

    def start_doc(self, attrs):
        self._need_chars = True

    def end_doc(self):
        content = self._get_chars_and_disable()
        if self._stack:
            peek = self._stack[-1]
            if isinstance(peek, _AbstractKeywordData):
                peek.doc = content

    def start_msg(self, attrs):
        # <msg timestamp="20221024 15:23:26.952" level="INFO">Some &lt;data &amp;encode &lt;/script&gt;</msg>
        self._need_chars = True
        self._curr_message_attrs = attrs

    def end_msg(self):
        self.send_delayed()
        content = self._get_chars_and_disable()

        level = self._curr_message_attrs["level"]
        timestamp = self._curr_message_attrs["timestamp"]
        html = self._curr_message_attrs.get("html")

        self._chars = []
        self._curr_message_attrs = None

        key = (level, content, timestamp)
        if key in self._found_logs:
            # RF duplicates messages at the end of the log. We just
            # want to show it once.
            return
        self._found_logs.add(key)

        self._listener.log_message(
            {
                "level": level,
                "message": content,
                "html": html,
                "timedelta": compute_timedelta(self._listener.initial_time, timestamp),
            }
        )

    def characters(self, content):
        if self._need_chars:
            self._chars.append(content)

    def start_status(self, attrs):
        self.send_delayed()
        if self._stack:
            self._stack[-1].status = _Status(attrs)

    def end_status(self):
        pass  # no-op


def parse_time(date_str):
    return datetime.datetime.strptime(date_str, "%Y%m%d %H:%M:%S.%f")


def convert_xml_to_rfstream(source, write: Optional[Callable[[str], None]] = None):
    """
    :param source:
        Either a string pointing to the path to be parsed or some stream-like
        object with the contents.

    :param write:
        A callable to be used to write the contents received (sent line-by-line).
    """
    from robot_out_stream import RFStream

    if write is None:

        def write(s):
            sys.stdout.write(s)

    def create_listener(robot_attrs):
        initial_date_str = robot_attrs["generated"]
        initial_time = parse_time(initial_date_str)
        kwargs = dict(
            __write__=write,
            __initial_time__=initial_time,
            __additional_info__=[f"Generated from output.xml"],
            __uuid__="gen-from-output-xml",
        )
        kwargs["--dir"] = "None"
        listener = RFStream(**kwargs)
        return listener

    xml.sax.parse(source, _XmlSaxParser(create_listener))
